"""
Models for file reading and writing.
"""

from .products import ProductsReader, ProductsWriter
from .receipt import ReceiptReader, ReceiptWriter

__all__ = ["ProductsReader", "ProductsWriter", "ReceiptReader", "ReceiptWriter"]



"""
Base type for measurable quantities and units.
"""

from decimal import Decimal
from typing import Callable, Generic, TypeVar, Union
import pint
from pint.facets.plain import PlainQuantity, PlainUnit
from typing_extensions import Self, TypeGuard

Dimension = Union[PlainQuantity[Decimal], PlainUnit]
DimensionT = TypeVar("DimensionT", bound=Dimension)
MeasurableT = TypeVar("MeasurableT", bound="Measurable")

UnitRegistry = pint.UnitRegistry(cache_folder=":auto:", non_int_type=Decimal)

class Measurable(Generic[DimensionT]):
    """
    A value that has operations to convert or derive it.
    """

    _wrappers: dict[type[Dimension], type["Measurable"]] = {}

    @classmethod
    def register_wrapper(cls, dimension: type[Dimension]) \
            -> Callable[[type["Measurable"]], type["Measurable"]]:
        """
        Register a measurable type which can wrap a `pint` dimension type.
        """

        def decorator(subclass: type["Measurable"]) -> type["Measurable"]:
            cls._wrappers[dimension] = subclass
            return subclass

        return decorator

    def __init__(self, value: DimensionT) -> None:
        super().__init__()
        self.value = value

    def _can_wrap(self, dimension: type[object]) -> TypeGuard[type[Dimension]]:
        return dimension in self._wrappers

    def _wrap(self, new: object) -> "Measurable":
        dimension = type(new)
        if self._can_wrap(dimension):
            return self._wrappers[dimension](new)

        raise TypeError("Could not convert to measurable object")

    @staticmethod
    def _unwrap(other: object) -> object:
        if isinstance(other, Measurable):
            return other.value
        return other

    def __lt__(self, other: object) -> bool:
        return self.value < self._unwrap(other)

    def __le__(self, other: object) -> bool:
        return self.value <= self._unwrap(other)

    def __eq__(self, other: object) -> bool:
        return self.value == self._unwrap(other)

    def __ne__(self, other: object) -> bool:
        return self.value != self._unwrap(other)

    def __gt__(self, other: object) -> bool:
        return self.value > self._unwrap(other)

    def __ge__(self, other: object) -> bool:
        return self.value >= self._unwrap(other)

    def __hash__(self) -> int:
        return hash(self.value)

    def __bool__(self) -> bool:
        return bool(self.value)

    def __mul__(self, other: object) -> "Measurable":
        return self._wrap(self.value * self._unwrap(other))

    def __truediv__(self: Self, other: object) -> "Measurable":
        return self._wrap(self.value / self._unwrap(other))

    __rmul__ = __mul__

    def __rtruediv__(self, other: object) -> "Measurable":
        return self._wrap(self._unwrap(other) / self.value)



"""
Unit type.
"""

from typing import Optional, Union
from pint.facets.plain import PlainUnit
from .base import Measurable, UnitRegistry

UnitNew = Optional[Union["Unit", PlainUnit, str]]

@Measurable.register_wrapper(UnitRegistry.Unit)
class Unit(Measurable[PlainUnit]):
    """
    A normalized unit value.
    """

    def __init__(self, unit: UnitNew) -> None:
        if isinstance(unit, Unit):
            unit = str(unit)
        elif unit is None:
            unit = ""
        super().__init__(UnitRegistry.Unit(unit))

    def __repr__(self) -> str:
        return f"Unit('{self.value!s}')"

    def __str__(self) -> str:
        return str(self.value)

    def __bool__(self) -> bool:
        return not self.value.dimensionless



"""
Abstract base classes for file reading, writing and parsing.
"""

from abc import ABCMeta
from collections.abc import Collection, Iterator
from datetime import datetime
import os
from pathlib import Path
import re
from typing import Any, Generic, IO, Optional, TypeVar
import yaml
from rechu.models.base import Base, GTIN, Price, Quantity

T = TypeVar('T', bound=Base)

class Reader(Generic[T], metaclass=ABCMeta):
    """
    File reader.
    """

    _mode = 'r'
    _encoding = 'utf-8'

    def __init__(self, path: Path, updated: datetime = datetime.min):
        self._path = path
        self._updated = updated

    @property
    def path(self) -> Path:
        """
        Retrieve the path from which to read the models.
        """

        return self._path

    def read(self) -> Iterator[T]:
        """
        Read the file from the path and yield specific models from it.
        """

        with self._path.open(self._mode, encoding=self._encoding) as file:
            yield from self.parse(file)

    def parse(self, file: IO) -> Iterator[T]:
        """
        Parse an open file and yield specific models from it.

        This method raises `TypeError` or subclasses if certain data in the
        file does not have the correct type, and `ValueError` or subclasses if
        the data has inconsistent or out-of-range values.
        """

        raise NotImplementedError('Must be implemented by subclasses')

class YAMLReader(Reader[T], metaclass=ABCMeta):
    """
    YAML file reader.
    """

    def load(self, file: IO) -> Any:
        """
        Load the YAML file as a Python value.
        """

        try:
            return yaml.safe_load(file)
        except yaml.parser.ParserError as error:
            raise TypeError(f"YAML failure in file '{self._path}' {error}") \
                from error

class Writer(Generic[T], metaclass=ABCMeta):
    """
    File writer.
    """

    _mode = 'w'
    _encoding = 'utf-8'

    def __init__(self, path: Path, models: Collection[T],
                 updated: Optional[datetime] = None):
        self._path = path
        self._models = models
        self._updated = updated

    @property
    def path(self) -> Path:
        """
        Retrieve the path to which to write the models.
        """

        return self._path

    def write(self) -> None:
        """
        Write the models to the path.
        """

        with self._path.open(self._mode, encoding=self._encoding) as file:
            self.serialize(file)

        if self._updated is not None:
            os.utime(self._path, times=(self._updated.timestamp(),
                                        self._updated.timestamp()))

    def serialize(self, file: IO) -> None:
        """
        Write a serialized variant of the models to the open file.
        """

        raise NotImplementedError('Must be implemented by subclasses')

class YAMLWriter(Writer[T], metaclass=ABCMeta):
    """
    YAML file writer.
    """

    TAG_INT = 'tag:yaml.org,2002:int'
    TAG_FLOAT = 'tag:yaml.org,2002:float'
    TAG_STR = 'tag:yaml.org,2002:str'

    @classmethod
    def _represent_gtin(cls, dumper: yaml.Dumper, data: GTIN) -> yaml.Node:
        return dumper.represent_scalar(cls.TAG_INT, f"{data:0>14}")

    @classmethod
    def _represent_price(cls, dumper: yaml.Dumper, data: Price) -> yaml.Node:
        return dumper.represent_scalar(cls.TAG_FLOAT, str(data))

    @classmethod
    def _represent_quantity(cls, dumper: yaml.Dumper,
                            data: Quantity) -> yaml.Node:
        if data.unit:
            return dumper.represent_scalar(cls.TAG_STR, str(data))
        return dumper.represent_scalar(cls.TAG_INT, str(int(data)))

    def save(self, data: Any, file: IO) -> None:
        """
        Save the YAML file from a Python value.
        """

        yaml.add_implicit_resolver(self.TAG_INT, re.compile(r'^\d{14}$'),
                                   list('0123456789'))
        yaml.add_representer(GTIN, self._represent_gtin)
        yaml.add_representer(Price, self._represent_price)
        yaml.add_representer(Quantity, self._represent_quantity)
        yaml.dump(data, file, width=80, indent=2, default_flow_style=None,
                  sort_keys=False)



"""
Receipt file handling.
"""

from collections.abc import Collection, Iterator
from datetime import datetime, date
from decimal import Decimal
from pathlib import Path
from typing import IO, Optional, Union
from typing_extensions import Required, TypedDict
from .base import YAMLReader, YAMLWriter
from ..models.base import Price, Quantity
from ..models.receipt import Discount, ProductItem, Receipt

_ProductItem = list[Union[str, float, Price, Quantity]]
_Discount = list[Union[str, Price]]

class _Receipt(TypedDict, total=False):
    date: Required[date]
    shop: Required[str]
    products: Required[list[_ProductItem]]
    bonus: list[_Discount]

class ReceiptReader(YAMLReader[Receipt]):
    """
    Receipt file reader.
    """

    def parse(self, file: IO) -> Iterator[Receipt]:
        data: _Receipt = self.load(file)
        if not isinstance(data, dict):
            raise TypeError(f"File '{self._path}' does not contain a mapping")
        try:
            receipt = Receipt(filename=self._path.name, updated=self._updated,
                              date=data['date'], shop=str(data['shop']))
            receipt.products = [
                self._product(position, item)
                for position, item in enumerate(data['products'])
            ]
            receipt.discounts = [
                self._discount(position, item, receipt.products)
                for position, item in enumerate(data.get('bonus', []))
            ]
        except KeyError as error:
            raise TypeError(f"Missing field in file '{self._path}': {error}") \
                from error
        yield receipt

    def _product(self, position: int, item: _ProductItem) -> ProductItem:
        if len(item) < 3:
            raise TypeError(f"Product item has too few elements: {len(item)}")
        quantity = Quantity(item[0])
        if not isinstance(item[2], (str, float, Decimal)):
            raise TypeError(f"Price '{item[2]!r}' could not be converted")
        discount_indicator = str(item[3]) if len(item) > 3 else None
        return ProductItem(quantity=quantity, label=str(item[1]),
                           price=Price(item[2]),
                           discount_indicator=discount_indicator,
                           position=position,
                           amount=quantity.amount, unit=quantity.unit)

    def _discount(self, position: int, item: _Discount,
                  products: list[ProductItem]) -> Discount:
        if len(item) < 2:
            raise TypeError(f"Discount has too few elements: {len(item)}")
        discount = Discount(label=str(item[0]), price_decrease=Price(item[1]),
                            position=position)
        seen = 0
        for label in item[2:]:
            for index, product in enumerate(products[seen:]):
                if product.discount_indicator and label == product.label:
                    discount.items.append(product)
                    seen += index + 1
                    break
        return discount

class ReceiptWriter(YAMLWriter[Receipt]):
    """
    Receipt file writer.
    """

    def __init__(self, path: Path, models: Collection[Receipt],
                 updated: Optional[datetime] = None):
        if not models or len(models) > 1:
            raise TypeError('Can only write exactly one receipt per file')
        self._model = next(iter(models))
        if updated is None:
            updated = self._model.updated
        super().__init__(path, models, updated=updated)

    @staticmethod
    def _get_product(product: ProductItem) -> _ProductItem:
        data: _ProductItem = [product.quantity, product.label, product.price]
        if product.discount_indicator is not None:
            data.append(product.discount_indicator)
        return data

    @staticmethod
    def _get_discount(discount: Discount) -> _Discount:
        data: list[Union[str, Price]] = [
            discount.label, discount.price_decrease
        ]
        data.extend([item.label for item in discount.items])
        return data

    def serialize(self, file: IO) -> None:
        data: _Receipt = {
            'date': self._model.date,
            'shop': self._model.shop,
            'products': [
                self._get_product(product) for product in self._model.products
            ],
            'bonus': [
                self._get_discount(bonus) for bonus in self._model.discounts
            ]
        }
        self.save(data, file)



"""
Base for receipt subcommands.
"""

from argparse import ArgumentParser, FileType, Namespace
import logging
import os
from pathlib import Path
from typing import Callable, Iterable, Generic, Optional, Sequence, TypeVar, \
    Union
from typing_extensions import TypedDict
from .. import __name__ as NAME, __version__ as VERSION
from ..settings import Settings

class SubparserKeywords(TypedDict, total=False):
    """
    Keyword arguments acceptable for subcommands to register to a subparser of
    an argument parser.
    """

    help: Optional[str]
    aliases: Sequence[str]
    description: Optional[str]
    epilog: Optional[str]
    prefix_chars: str
    fromfile_prefix_chars: Optional[str]
    add_help: bool
    allow_abbrev: bool

ArgumentT = TypeVar('ArgumentT')

class ArgumentKeywords(Generic[ArgumentT], TypedDict, total=False):
    """
    Keyword arguments acceptable for registering an argument to a subparser of
    an argument parser.
    """

    action: str
    nargs: Optional[Union[int, str]]
    const: ArgumentT
    default: Union[ArgumentT, str]
    type: Union[Callable[[str], ArgumentT], FileType]
    choices: Optional[Iterable[ArgumentT]]
    required: bool
    help: Optional[str]
    metavar: Optional[Union[str, tuple[str, ...]]]
    dest: Optional[str]

ArgumentSpec = tuple[Union[str, tuple[str, ...]], ArgumentKeywords]
SubparserArguments = Iterable[ArgumentSpec]

class _SubcommandHolder(Namespace): # pylint: disable=too-few-public-methods
    subcommand: str = ''
    log: str = 'INFO'

class Base(Namespace):
    """
    Abstract command handling.
    """

    _commands: dict[str, type['Base']] = {}
    program: str = NAME
    subcommand: str = ''
    subparser_keywords: SubparserKeywords = {}
    subparser_arguments: SubparserArguments = []

    @classmethod
    def register(cls, name: str) -> Callable[[type['Base']], type['Base']]:
        """
        Register a subcommand.
        """

        def decorator(subclass: type['Base']) -> type['Base']:
            cls._commands[name] = subclass
            subclass.subcommand = name
            return subclass

        return decorator

    @classmethod
    def get_command(cls, name: str) -> 'Base':
        """
        Create a command instance for the given subcommand name.
        """

        return cls._commands[name]()

    @classmethod
    def register_arguments(cls) -> ArgumentParser:
        """
        Create an argument parser for all registered subcommands.
        """

        parser = ArgumentParser(prog=cls.program,
                                description='Receipt cataloging hub')
        parser.add_argument('--version', action='version',
                            version=f'{NAME} {VERSION}')
        parser.add_argument('--log',
                            choices=[
                                "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"
                            ],
                            default="INFO", help='Log level')
        subparsers = parser.add_subparsers(dest='subcommand',
                                           help='Subcommands')
        for name, subclass in cls._commands.items():
            subparser = subparsers.add_parser(name,
                                              **subclass.subparser_keywords)
            for (argument, keywords) in subclass.subparser_arguments:
                if isinstance(argument, str):
                    subparser.add_argument(argument, **keywords)
                else:
                    subparser.add_argument(*argument, **keywords)

        return parser

    @classmethod
    def start(cls, executable: str, argv: Sequence[str]) -> None:
        """
        Parse arguments from a sequence of command line arguments and determine
        which command to run, register any arguments to it and finally execute
        the action of the command.
        """

        cls.program = Path(argv[0]).name
        if cls.program == "__main__.py":
            python = Path(executable)
            if str(python.parent) in os.get_exec_path():
                executable = python.name
            cls.program = f"{executable} -m {NAME}"

        parser = cls.register_arguments()
        if len(argv) <= 1:
            parser.print_usage()
            return

        holder = _SubcommandHolder()
        parser.parse_known_args(argv[1:], namespace=holder)

        logging.getLogger(NAME).setLevel(getattr(logging, holder.log, 0))

        command = cls.get_command(holder.subcommand)
        command.program = cls.program
        command.subcommand = holder.subcommand
        parser.parse_args(argv[1:], namespace=command)
        command.run()

    def __init__(self) -> None:
        super().__init__()
        self.settings = Settings.get_settings()
        self.logger = logging.getLogger(self.__class__.__module__)

    def run(self) -> None:
        """
        Execute the command.
        """

        raise NotImplementedError('Must be implemented by subclasses')



"""
Bag of files containing multiple grouped models that share common properties.
"""

from collections.abc import Iterable, Iterator, Sequence
from pathlib import Path
from typing import Mapping, Optional, TypeVar
from sqlalchemy.orm import Session
from ..io.base import Writer
from ..models.base import Base as ModelBase

T = TypeVar('T', bound=ModelBase)

Selectors = list[dict[str, Optional[str]]]

class Inventory(Mapping[Path, Sequence[T]]):
    """
    An inventory of a type of model grouped by one or more characteristics,
    which are concretized in file names.
    """

    @classmethod
    def spread(cls, models: Iterable[T]) -> "Inventory[T]":
        """
        Create an inventory based on provided models by assigning them to groups
        that each belongs to.
        """

        raise NotImplementedError('Spreading must be implemented by subclass')

    @classmethod
    def select(cls, session: Session,
               selectors: Optional[Selectors] = None) -> "Inventory[T]":
        """
        Create an inventory based on models stored in the database.
        """

        raise NotImplementedError('Selection must be implemented by subclass')

    @classmethod
    def read(cls) -> "Inventory[T]":
        """
        Create an inventory based on models stored in files.
        """

        raise NotImplementedError('Reading must be implemented by subclass')

    def get_writers(self) -> Iterator[Writer[T]]:
        """
        Obtain writers for each inventory file.
        """

        raise NotImplementedError('Writers must be implemented by subclass')

    def write(self) -> None:
        """
        Write an inventory to files.
        """

        for writer in self.get_writers():
            writer.write()

    def merge_update(self, other: "Inventory[T]", update: bool = True,
                     only_new: bool = False) -> "Inventory[T]":
        """
        Find groups with models that are added or updated in the other inventory
        compared to the current inventory. The returned inventory contains the
        new, existing and merged models grouped by path; only paths with changes
        are included. The products in the current inventory are updated as well.
        If `update` is enabled, then new models are added to and changed models
        updated in the current inventory; this is the default. If `update` is
        disabled, then only the updated models are provided in the return value
        and the current object also remains immutable. If `only_new` is enabled,
        then models that existed but had changes are not considered, just like
        unchanged models; `only_new` inherently disables `update`.
        """

        raise NotImplementedError('Merging must be implemented by subclass')



"""
Receipt cataloging module.
"""

from logging.config import dictConfig

__all__: list[str] = []
__version__ = "0.0.1"

dictConfig({
    'version': 1,
    'formatters': {
        'generic_dated': {
            'format': '%(asctime)s %(levelname)-5.5s [%(name)s] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'stderr': {
            'class': 'logging.StreamHandler',
            'formatter': 'generic_dated',
            'stream': 'ext://sys.stderr',
            'level': 'NOTSET',
        }
    },
    'loggers': {
        __name__: {
            'level': 'NOTSET',
            'handlers': ['stderr']
        }
    }
})



"""
Add product relation and shop

Revision ID: d6854aeb9756
Revises: 9d5f6a8f4944
Create Date: 2025-04-23 23:21:27.451326
"""
# pylint: disable=invalid-name

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column
from rechu.inventory.products import Products


# Revision identifiers, used by Alembic.
revision: str = 'd6854aeb9756'
down_revision: Union[str, None] = '9d5f6a8f4944'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """
    Perform the upgrade.
    """

    with op.batch_alter_table('product', schema=None) as batch_op:
        batch_op.add_column(sa.Column('shop', sa.String(length=32),
                                      nullable=True))
    with op.batch_alter_table('product', schema=None) as batch_op:
        product = table('product', column('id', sa.Integer()),
                        column('shop', sa.String(length=32)))
        start = 0
        for products in Products.read().values():
            end = start + len(products)
            end += sum(len(product.range) for product in products)
            batch_op.execute(product.update().where(product.c.id.between(start,
                                                                         end))
                             .values(shop=products[0].shop))
            start = end
        batch_op.alter_column('shop',
                              existing_type=sa.String(length=32),
                              existing_nullable=True,
                              nullable=False)

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        batch_op.add_column(sa.Column('product_id', sa.Integer(),
                                      nullable=True))
        receipt_product_product_key = 'fk_receipt_product_product_id_product'
        batch_op.create_foreign_key(batch_op.f(receipt_product_product_key),
                                    'product', ['product_id'], ['id'],
                                    ondelete='SET NULL')

    # ### end Alembic commands ###


def downgrade() -> None:
    """
    Perform the downgrade.
    """

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        receipt_product_product_key = 'fk_receipt_product_product_id_product'
        batch_op.drop_constraint(batch_op.f(receipt_product_product_key),
                                 type_='foreignkey')
        batch_op.drop_column('product_id')

    with op.batch_alter_table('product', schema=None) as batch_op:
        batch_op.drop_column('shop')

    # ### end Alembic commands ###



"""
Add receipt item extracted quantity fields

Revision ID: 8ef12eb24650
Revises: b1a7a91c8de8
Create Date: 2025-05-29 22:55:58.519845
"""
# pylint: disable=invalid-name

from typing import Sequence, Union

from alembic import context, op
from alembic.operations import BatchOperations
import sqlalchemy as sa
from sqlalchemy.sql import TableClause, table, column
from rechu.database import Database
from rechu.models.base import Quantity

# Revision identifiers, used by Alembic.
revision: str = '8ef12eb24650'
down_revision: Union[str, None] = 'b1a7a91c8de8'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

Amounts = list[dict[str, Union[float, str]]]
Units = list[dict[str, Union[int, str]]]

def upgrade() -> None:
    """
    Perform the upgrade.
    """

    product = table('receipt_product', column('id', sa.Integer()),
                    column('quantity', sa.String()),
                    column('amount', sa.Float()),
                    column('unit', sa.String()))
    amounts, units = collect_quantity(product)

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        batch_op.add_column(sa.Column('amount', sa.Float(), nullable=True))

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        update_amount(batch_op, product, amounts)

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        batch_op.alter_column('amount',
                              existing_type=sa.Float(),
                              existing_nullable=True,
                              nullable=False)
        batch_op.add_column(sa.Column('unit', sa.String(), nullable=True))

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        update_unit(batch_op, product, units)

    # ### end Alembic commands ###

def collect_quantity(product: TableClause) -> tuple[Amounts, Units]:
    """
    Extract quantity fields.
    """

    amounts = []
    units = []
    with Database() as session:
        if context.is_offline_mode():
            connection = session.connection()
        else:
            connection = op.get_bind()
        for row in connection.execute(sa.select(product.c.id,
                                                product.c.quantity)):
            quantity = Quantity(row.quantity)
            amounts.append({
                "row_id": row.id,
                "amount": quantity.amount
            })
            if quantity.unit is not None:
                units.append({
                    "row_id": row.id,
                    "unit": str(quantity.unit)
                })

    return amounts, units

def update_amount(batch_op: BatchOperations, product: TableClause,
                  amounts: Amounts) -> None:
    """
    Update the amount column.
    """

    for amount in amounts:
        batch_op.execute(product.update()
                         .where(product.c.id == amount["row_id"])
                         .values({"amount": amount["amount"]}))

def update_unit(batch_op: BatchOperations, product: TableClause,
                units: Units) -> None:
    """
    Update the unit column.
    """

    for unit in units:
        batch_op.execute(product.update()
                         .where(product.c.id == unit["row_id"])
                         .values({"unit": unit["unit"]}))

def downgrade() -> None:
    """
    Perform the downgrade.
    """

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        batch_op.drop_column('unit')
        batch_op.drop_column('amount')

    # ### end Alembic commands ###



"""
Quantity type.
"""

from decimal import Decimal
from typing import Optional, Union
from pint.errors import UndefinedUnitError
from pint.facets.plain import PlainQuantity
from typing_extensions import Self
from .base import Measurable, UnitRegistry
from .unit import Unit, UnitNew

QuantityNew = Union["Quantity", PlainQuantity, Decimal, float, str]

@Measurable.register_wrapper(UnitRegistry.Quantity)
class Quantity(Measurable[PlainQuantity]):
    """
    A quantity value with an optional dimension with original input preserved.
    """

    def __init__(self, value: QuantityNew, unit: UnitNew = None) -> None:
        if isinstance(value, Quantity):
            value = str(value)
        elif isinstance(value, PlainQuantity) and value.dimensionless:
            value = value.magnitude
        if isinstance(unit, Unit):
            unit = str(unit)
        try:
            super().__init__(UnitRegistry.Quantity(value, units=unit))
        except UndefinedUnitError as error:
            raise ValueError("Could not create a quantity with unit") from error
        if unit is None or self.value.dimensionless:
            self._original = str(value)
        else:
            self._original = f"{value}{unit}"

    @property
    def amount(self) -> float:
        """
        Retrieve the magnitude of the quantity as a plain number.
        """

        return self.value.magnitude

    @property
    def unit(self) -> Optional[Unit]:
        """
        Retrieve the normalized unit of the quantity, or `None` if it has no
        unit dimensionality.
        """

        if self.value.dimensionless:
            return None

        return Unit(self.value.units)

    def __repr__(self) -> str:
        if self.value.dimensionless:
            return f"Quantity({self._original!r})"

        return f"Quantity('{self.amount!s}', '{self.value.units!s}')"

    def __str__(self) -> str:
        return self._original

    def __int__(self) -> int:
        return int(self.amount)

    def __float__(self) -> float:
        return float(self.amount)

    def __add__(self: Self, other: object) -> Self:
        return self.__class__(self.value + self._unwrap(other))

    def __sub__(self: Self, other: object) -> Self:
        return self.__class__(self.value - self._unwrap(other))

    def __floordiv__(self: Self, other: object) -> Self:
        return self.__class__(self.value // self._unwrap(other))

    def __mod__(self: Self, other: object) -> Self:
        return self.__class__(self.value % self._unwrap(other))

    def __pow__(self: Self, other: object) -> Self:
        return self.__class__(self.value ** self._unwrap(other))

    __radd__ = __add__

    def __rsub__(self: Self, other: object) -> Self:
        return self.__class__(self._unwrap(other) - self.value)

    def __rfloordiv__(self: Self, other: object) -> Self:
        return self.__class__(self._unwrap(other) // self.value)

    def __rmod__(self: Self, other: object) -> Self:
        return self.__class__(self._unwrap(other) % self.value)

    def __rpow__(self: Self, other: object) -> Self:
        return self.__class__(self._unwrap(other) ** self.value)

    def __neg__(self: Self) -> Self:
        return self.__class__(-self.value)

    def __pos__(self: Self) -> Self:
        return self.__class__(+self.value)

    def __abs__(self) -> Self:
        return self.__class__(abs(self.value))

    def __round__(self: Self, ndigits: Optional[int] = 0) -> Self:
        return self.__class__(round(self.value, ndigits=ndigits))



"""
Type decorators for model type annotation maps.
"""

from typing import Any, Generic, Optional, Protocol, TypeVar
from sqlalchemy.engine import Dialect
from sqlalchemy.types import String, TypeDecorator, TypeEngine
from typing_extensions import Self

class Convertible(Protocol):
    # pylint: disable=too-few-public-methods
    """
    A type which can be created from another input type.
    """

    def __new__(cls: type[Self], value: object) -> Self:
        ...

T = TypeVar('T', bound=Convertible)
ST = TypeVar('ST', bound=Convertible)

class SerializableType(TypeDecorator[T], Generic[T, ST]):
    # pylint: disable=too-many-ancestors
    """
    Type decoration handler for attributes.
    """

    # Default implementation
    impl: TypeEngine = String()

    def process_literal_param(self, value: Optional[T],
                              dialect: Dialect) -> str:
        if value is None:
            return "NULL"
        processor = self.impl.literal_processor(dialect)
        if processor is None: # pragma: no cover
            raise TypeError("There should be a literal processor for SQL type")
        return processor(self.serialized_type(value))

    def process_bind_param(self, value: Optional[T], dialect: Dialect) -> Any:
        if value is None:
            return None
        return self.serialized_type(value)

    def process_result_value(self, value: Optional[Any],
                             dialect: Dialect) -> Optional[T]:
        if value is None:
            return None
        return self.serializable_type(value)

    @property
    def python_type(self) -> type[Any]:
        return self.serializable_type

    @property
    def serializable_type(self) -> type[T]:
        """
        Retrieve the type to use for result values of this serialized type.
        """

        raise NotImplementedError("Must be implemented by subclasses")

    @property
    def serialized_type(self) -> type[ST]:
        """
        Retrieve the type to use for storing the values in the database.
        """

        raise NotImplementedError("Must be implemented by subclasses")



"""
Base model for receipt cataloging.
"""

from sqlalchemy import MetaData
from sqlalchemy.orm import DeclarativeBase, registry
from ..types.measurable import Quantity, Unit, QuantityType, UnitType
from ..types.quantized import GTIN, Price, GTINType, PriceType


class Base(DeclarativeBase): # pylint: disable=too-few-public-methods
    """
    Base ORM model class for receipt models.
    """

    metadata = MetaData(naming_convention={
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    })

    registry = registry(type_annotation_map={
        Price: PriceType,
        Quantity: QuantityType,
        Unit: UnitType,
        GTIN: GTINType
    })



"""
Type decorators for measurable types.
"""

from sqlalchemy import String
from .quantity import Quantity
from .unit import Unit
from ..decorator import SerializableType

class QuantityType(SerializableType[Quantity, str]):
    # pylint: disable=too-many-ancestors
    """
    Type decoration handler for quantities.
    """

    cache_ok = True
    impl = String()

    @property
    def serializable_type(self) -> type[Quantity]:
        return Quantity

    @property
    def serialized_type(self) -> type[str]:
        return str

class UnitType(SerializableType[Unit, str]):
    # pylint: disable=too-many-ancestors
    """
    Type decoration handler for units.
    """

    cache_ok = True
    impl = String()

    @property
    def serializable_type(self) -> type[Unit]:
        return Unit

    @property
    def serialized_type(self) -> type[str]:
        return str



"""
Drop foreign key for receipt shop

Revision ID: 9d5f6a8f4944
Revises: 24c54f418b92
Create Date: 2025-02-10 20:23:02.768133
"""
# pylint: disable=invalid-name

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import column, table

# Revision identifiers, used by Alembic.
revision: str = '9d5f6a8f4944'
down_revision: Union[str, None] = '24c54f418b92'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """
    Perform the upgrade.
    """

    with op.batch_alter_table('receipt', schema=None) as batch_op:
        batch_op.drop_constraint('fk_receipt_shop_shop', type_='foreignkey')

    # ### end Alembic commands ###


def downgrade() -> None:
    """
    Perform the downgrade.
    """

    with op.batch_alter_table('receipt', schema=None) as batch_op:
        receipt = table('receipt', column('shop', sa.String(32)))
        shop = table('shop', column('key', sa.String(32)))
        select = sa.select(receipt.c.shop.label("key")).distinct()
        batch_op.execute(shop.insert().from_select(["key"], select))
        batch_op.create_foreign_key('fk_receipt_shop_shop', 'shop', ['shop'],
                                    ['key'])

    # ### end Alembic commands ###



"""
Steps for creating a receipt in new subcommand.
"""

from datetime import date
from itertools import chain
import logging
import os
from pathlib import Path
import re
import shutil
import subprocess
import sys
import tempfile
from typing import Optional, Union
from sqlalchemy import select
from sqlalchemy.sql.functions import count, min as min_
from typing_extensions import Required, TypedDict
from .input import Input, InputSource
from ...database import Database
from ...inventory.products import Products as ProductInventory
from ...io.products import ProductsReader, ProductsWriter, IDENTIFIER_FIELDS, \
    OPTIONAL_FIELDS
from ...io.receipt import ReceiptReader, ReceiptWriter
from ...matcher.product import ProductMatcher, MapKey
from ...models.base import Base as ModelBase, GTIN, Price, Quantity
from ...models.product import Product, LabelMatch, PriceMatch, DiscountMatch
from ...models.receipt import Discount, ProductItem, Receipt

Menu = dict[str, 'Step']
ProductsMeta = set[Product]

LOGGER = logging.getLogger(__name__)

class _Matcher(TypedDict, total=False):
    model: Required[type[ModelBase]]
    key: Required[str]
    extra_key: str
    input_type: type[Input]
    options: Optional[str]
    normalize: str

class ResultMeta(TypedDict, total=False):
    """
    Result of a step being run, indicator additional metadata to update.

    - 'receipt_path': Boolean indicating pdate the path of the receipt based on
      receipt metadata.
    """

    receipt_path: bool

_MetaResult = tuple[bool, Optional[str], bool]
_Pairs = tuple[tuple[Product, ProductItem], ...]

class ReturnToMenu(RuntimeError):
    """
    Indication that the step is interrupted to return to a menu.
    """

    def __init__(self, msg: str = '') -> None:
        super().__init__(msg)
        self.msg = msg

class Step:
    """
    Abstract base class for a step during receipt creation.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource) -> None:
        self._receipt = receipt
        self._input = input_source

    def run(self) -> ResultMeta:
        """
        Perform the step. Returns whether there is additional metadata which
        needs to be updated outside of the step.
        """

        raise NotImplementedError('Step must be implemented by subclasses')

    @property
    def description(self) -> str:
        """
        Usage message that explains what the step does.
        """

        raise NotImplementedError('Description must be implemented by subclass')

    @property
    def final(self) -> bool:
        """
        Whether this step finalizes the receipt generation.
        """

        return False

class Read(Step):
    """
    Step to check if there are any new or updated product metadata entries in
    the file inventory that should be synchronized with the database inventory
    before creating and matching receipt products.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource,
                 matcher: ProductMatcher) -> None:
        super().__init__(receipt, input_source)
        self._matcher = matcher

    def run(self) -> ResultMeta:
        with Database() as session:
            session.expire_on_commit = False
            database = ProductInventory.select(session)
            self._matcher.fill_map(database)
            files = ProductInventory.read()
            updates = database.merge_update(files, update=False)
            deleted = files.merge_update(database, update=False, only_new=True)
            paths = set(chain((path.name for path in updates.keys()),
                              (path.name for path in deleted.keys())))
            confirm = ''
            while paths and confirm != 'y':
                LOGGER.warning('Updated products files detected: %s', paths)
                confirm = self._input.get_input('Confirm reading products (y)',
                                                str)

            for group in updates.values():
                for product in group:
                    self._matcher.add_map(product)
                    session.merge(product)
            for group in deleted.values():
                for product in group:
                    LOGGER.warning('Deleting %r', product)
                    self._matcher.discard_map(product)
                    session.delete(product)

            for key in ('brand', 'category', 'type'):
                field = getattr(Product, key)
                self._input.update_suggestions({
                    f'{key}s': list(session.scalars(select(field).distinct()
                                                    .filter(field.is_not(None))
                                                    .order_by(field)))
                })

        return {}

    @property
    def description(self) -> str:
        return "Check updated receipt metadata YAML files"

class Products(Step):
    """
    Step to add products.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource,
                 matcher: ProductMatcher, products: ProductsMeta) -> None:
        super().__init__(receipt, input_source)
        self._matcher = matcher
        self._products = products

    def run(self) -> ResultMeta:
        self._matcher.discounts = bool(self._receipt.discounts)
        ok = True
        while ok:
            ok = self.add_product()

        return {}

    def add_product(self) -> bool:
        """
        Request fields for a product and add it to the receipt.
        """

        prompt = 'Quantity (empty or 0 to end products, ? to menu, ! cancels)'
        if self._receipt.products:
            previous = self._receipt.products[-1]
            # Check if the previous product item has a product metadata match
            # If not, we might want to create one right now
            with Database() as session:
                pairs = tuple(self._matcher.find_candidates(session, (previous,),
                                                            self._products))
                dedupe = tuple(self._matcher.filter_duplicate_candidates(pairs))
                amount = self._make_meta(previous, prompt, pairs, dedupe)
        else:
            amount = self._input.get_input(prompt, str)

        if amount in {'', '0'}:
            return False
        if amount == '?':
            raise ReturnToMenu
        if amount == '!':
            LOGGER.info('Removing previous product: %r',
                        self._receipt.products[-1:])
            self._receipt.products[-1:] = []
            return True

        try:
            quantity = Quantity(amount)
        except ValueError as error:
            LOGGER.error("Could not validate quantity: %s", error)
            return True

        label = self._input.get_input('Label', str, options='products')

        with Database() as session:
            self._input.update_suggestions({'prices': [
                str(price)
                for price in session.scalars(select(ProductItem.price, count())
                                     .where(ProductItem.label == label)
                                     .group_by(ProductItem.price)
                                     .order_by(count()))
            ]})
        price = self._input.get_input('Price', Price, options='prices')

        discount = self._input.get_input('Discount indicator', str)
        position = len(self._receipt.products)
        self._receipt.products.append(ProductItem(quantity=quantity,
                                                  label=label,
                                                  price=price,
                                                  discount_indicator=discount \
                                                      if discount != '' \
                                                      else None,
                                                  position=position,
                                                  amount=quantity.amount,
                                                  unit=quantity.unit))
        return True

    def _make_meta(self, item: ProductItem, prompt: str,
                   pairs: _Pairs, dedupe: _Pairs) -> Union[str, Quantity]:
        if dedupe and dedupe[0][0].discounts:
            LOGGER.info('Matched with %r excluding discounts', dedupe[0][0])
        elif dedupe:
            LOGGER.info('Matched with %r', dedupe[0][0])
        elif len(pairs) > 1:
            LOGGER.warning('Multiple metadata matches, ignoring for now')
        else:
            match = False
            while not match:
                meta_prompt = f'No metadata yet. Next {prompt.lower()} or key'
                key = self._input.get_input(meta_prompt, str, options='meta')
                if key in {'', '?', '!'} or key[0].isnumeric():
                    # Quantity or other product item command
                    return key

                product = ProductMeta(self._receipt, self._input,
                                      matcher=self._matcher,
                                      products=self._products)
                match = not product.add_product(item=item, initial_key=key)[0]

        return self._input.get_input(prompt, str)

    @property
    def description(self) -> str:
        return "Add products to receipt"

class Discounts(Step):
    """
    Step to add discounts.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource,
                 matcher: ProductMatcher) -> None:
        super().__init__(receipt, input_source)
        self._matcher = matcher

    def run(self) -> ResultMeta:
        ok = True
        self._matcher.discounts = True
        self._input.update_suggestions({
            'discount_items': sorted({product.label
                                      for product in self._receipt.products
                                      if product.discount_indicator})
        })
        while ok:
            ok = self.add_discount()

        return {}

    def add_discount(self) -> bool:
        """
        Request fields and items for a discount and add it to the receipt.
        """

        prompt = 'Discount label (empty to end discounts, ! cancels)'
        bonus = self._input.get_input(prompt, str, options='discounts')
        if bonus == '':
            return False
        if bonus == '?':
            raise ReturnToMenu
        if bonus == '!':
            if self._receipt.discounts:
                LOGGER.info('Removing previous discount: %r',
                            self._receipt.discounts[-1])
                self._receipt.discounts[-1].items = []
                self._receipt.discounts.pop()
            return True
        price = self._input.get_input('Price decrease (positive cancels)',
                                      Price)
        if price > 0:
            return True
        discount = Discount(label=bonus, price_decrease=price,
                            position=len(self._receipt.discounts))
        seen = 0
        try:
            while 0 <= seen < len(self._receipt.products):
                seen = self.add_discount_item(discount, seen)
        finally:
            if seen >= 0:
                self._receipt.discounts.append(discount)

        return True

    def add_discount_item(self, discount: Discount, seen: int) -> int:
        """
        Request fields for a discount item.
        """

        label = self._input.get_input('Product (in order on receipt, empty to '
                                      f'end "{discount.label}", ? to menu, ! '
                                      'cancels)', str, options='discount_items')
        if label == '':
            return sys.maxsize
        if label == '?':
            raise ReturnToMenu
        if label == '!':
            return -1
        discount_item: Optional[ProductItem] = None
        for index, product in enumerate(self._receipt.products[seen:]):
            if product.discount_indicator and label == product.label:
                discount_item = product
                discount.items.append(product)
                seen += index + 1
                break
        if discount_item is None:
            LOGGER.warning('No discounted product "%s" from #%d (%r)',
                           label, seen + 1, self._receipt.products[seen:])

        return seen

    @property
    def description(self) -> str:
        return "Add discounts to receipt"

class ProductMeta(Step):
    """
    Step to add product metadata that matches one or more products.
    """

    CONFIRM_ID = re.compile(r"^-?\d+$", re.ASCII)

    # Product metadata match entities
    MATCHERS: dict[str, _Matcher] = {
        'label': {
            'model': LabelMatch,
            'key': 'name',
            'options': 'products'
        },
        'price': {
            'model': PriceMatch,
            'key': 'value',
            'extra_key': 'indicator',
            'input_type': Price,
            'options': 'prices',
            'normalize': 'quantity'
        },
        'discount': {
            'model': DiscountMatch,
            'key': 'label',
            'options': 'discounts'
        }
    }

    def __init__(self, receipt: Receipt, input_source: InputSource,
                 matcher: ProductMatcher, products: ProductsMeta) -> None:
        super().__init__(receipt, input_source)
        self._matcher = matcher
        self._products = products

    def run(self) -> ResultMeta:
        ok = True
        initial_key: Optional[str] = None

        if not self._receipt.products:
            return {}

        # Check if there are any unmatched products on the receipt
        with Database() as session:
            candidates = self._matcher.find_candidates(session,
                                                       self._receipt.products,
                                                       self._products)
            pairs = self._matcher.filter_duplicate_candidates(candidates)
            matched_items = {item for _, item in pairs}
            LOGGER.info('%d/%d items already matched on receipt',
                        len(matched_items), len(self._receipt.products))

            if len(matched_items) == len(self._receipt.products):
                return {}

            min_date = session.scalar(select(min_(Receipt.date)))
            if min_date is None:
                min_date = self._receipt.date
            years = range(min_date.year, date.today().year + 1)
            self._input.update_suggestions({
                'indicators': [str(year) for year in years] + [
                    ProductMatcher.IND_MINIMUM, ProductMatcher.IND_MAXIMUM
                ] + [
                    str(product.unit) for product in self._receipt.products
                    if product.unit is not None
                ],
                'prices': [
                    str(product.price) for product in self._receipt.products
                ]
            })

        while (ok or initial_key == '!') and initial_key != '0' and \
            any(item not in matched_items for item in self._receipt.products):
            ok, initial_key = self.add_product(initial_key=initial_key,
                                               matched_items=matched_items)

        return {}

    def add_product(self, item: Optional[ProductItem] = None,
                    initial_key: Optional[str] = None,
                    matched_items: Optional[set[ProductItem]] = None) \
            -> tuple[bool, Optional[str]]:
        """
        Request fields for a product's metadata and add it to the database as
        well as a products YAML file. `item` is an optional product item
        from the receipt to specifically match the metadata for. `initial_key`
        is a metadata key to use for the first prompt. Returns whether to no
        longer attempt to create product metadata and the current prompt answer.
        """

        product = Product(shop=self._receipt.shop)

        matched, initial_key = self._fill_product(product, item=item,
                                                  initial_key=initial_key,
                                                  changed=False)
        while not matched:
            if initial_key in {'0', '!'}:
                return False, initial_key

            LOGGER.warning('Product %r does not match receipt item', product)
            changed = Product(shop=self._receipt.shop).merge(product)
            if not changed:
                return False, initial_key
            initial_key = self._get_key(product, item=item,
                                        initial_changed=changed)
            if initial_key == '':
                return changed, initial_key
            if initial_key in {'0', '!'}:
                return False, initial_key
            if initial_key == '?':
                raise ReturnToMenu

            matched, initial_key = self._fill_product(product, item=item,
                                                      initial_key=initial_key,
                                                      changed=changed)

        # Track product for later session merge and export
        LOGGER.info('Product created: %r', product)
        if product.generic is None:
            self._products.add(product)
            self._matcher.add_map(product)
        if matched_items is not None:
            matched_items.update(matched)

        return item is None, initial_key

    def _fill_product(self, product: Product,
                      item: Optional[ProductItem] = None,
                      initial_key: Optional[str] = None,
                      changed: bool = False) \
            -> tuple[set[ProductItem], Optional[str]]:
        initial_key = self._set_values(product, item=item,
                                       initial_key=initial_key, changed=changed)
        if initial_key == '':
            # Canceled creation/merged with already-matched product
            return set(), initial_key

        items = self._receipt.products if item is None else [item]
        matched = {item for item in items if item.product == product}
        with Database() as session:
            pairs = self._matcher.find_candidates(session, items,
                                                  self._products | {product})
            match_products = {product, product.generic}.union(product.range)
            match_products.discard(None)
            for meta, match in self._matcher.filter_duplicate_candidates(pairs):
                if meta in match_products:
                    matched.add(match)
                    if not match.discounts and product.discounts:
                        LOGGER.info('Matched with %r excluding discounts',
                                    match)
                    else:
                        LOGGER.info('Matched with item: %r', match)
                        match.product = product

        return matched, initial_key

    def _set_values(self, product: Product, item: Optional[ProductItem] = None,
                    initial_key: Optional[str] = None,
                    changed: bool = False) -> Optional[str]:
        ok = True
        while ok:
            ok, initial_key, changed = \
                self._add_key_value(product, item=item, initial_key=initial_key,
                                    initial_changed=changed)

        return initial_key

    def _add_key_value(self, product: Product,
                       item: Optional[ProductItem] = None,
                       initial_key: Optional[str] = None,
                       initial_changed: Optional[bool] = None) -> _MetaResult:
        key = self._get_key(product, item=item, initial_key=initial_key,
                            initial_changed=initial_changed)

        if key == 'range':
            return self._set_range(product, item, initial_changed)
        if key == 'view':
            return self._view(product, item, initial_changed)
        if key == 'edit':
            return self._edit(product, item, initial_changed)
        if key in {'', '0', '!'}:
            return False, key if key == '0' else None, bool(initial_changed)
        if key == '?':
            raise ReturnToMenu

        try:
            value = self._get_value(product, item, key)
        except KeyError:
            LOGGER.warning('Unrecognized metadata key %s', key)
            return True, None, bool(initial_changed)

        self._set_key_value(product, item, key, value)

        # Check if product matchers/identifiers clash
        return self._check_duplicate(product)

    @staticmethod
    def _get_initial_range(product: Product) -> Product:
        initial = product.copy()
        initial.range = []
        for field in IDENTIFIER_FIELDS:
            setattr(initial, field, None)
        return initial

    def _set_range(self, product: Product,
                   item: Optional[ProductItem],
                   initial_changed: Optional[bool] = None) -> _MetaResult:
        if product.generic is not None:
            LOGGER.warning('Cannot add product range to non-generic product')
            return True, None, bool(initial_changed)

        initial = self._get_initial_range(product)
        product_range = initial.copy()
        product_range.generic = product
        initial_key = self._set_values(product_range, item=item)
        if initial_key == '' or not initial.merge(product_range):
            product_range.generic = None
            return True, None, bool(initial_changed)

        return True, initial_key, True

    def _view(self, product: Product, item: Optional[ProductItem],
              initial_changed: Optional[bool] = None) -> _MetaResult:
        if item is not None:
            LOGGER.info('Receipt product item to match: %r', item)
        else:
            View(self._receipt, self._input, products=self._products).run()

        output = self._input.get_output()
        print(file=output)
        print('Current product metadata draft:', file=output)
        ProductsWriter(Path("products.yml"), (product,),
                       shared_fields=()).serialize(output)

        if initial_changed:
            return self._check_duplicate(product)
        return True, None, False

    def _edit(self, product: Product, item: Optional[ProductItem],
              initial_changed: Optional[bool] = None) -> _MetaResult:
        with tempfile.NamedTemporaryFile('w', suffix='.yml') as tmp_file:
            tmp_path = Path(tmp_file.name)
            editable = product if product.generic is None else product.generic
            writer = ProductsWriter(tmp_path, (editable,), shared_fields=())
            writer.write()
            if item is not None:
                tmp_file.write(f'# Product to match: {item!r}')

            edit = Edit(self._receipt, self._input)
            edit.execute_editor(tmp_file.name)

            reader = ProductsReader(tmp_path)
            try:
                new_product = next(reader.read())
                if product.generic is not None:
                    range_index = editable.range.index(product)
                    editable.replace(new_product)
                    product.replace(editable.range[range_index])
                    editable.range[range_index] = product
                else:
                    product.replace(new_product)
            except (StopIteration, TypeError, ValueError, IndexError):
                LOGGER.exception('Invalid or missing edited product YAML')
                return True, None, bool(initial_changed)

        return self._check_duplicate(product)

    def _get_key(self, product: Product, item: Optional[ProductItem] = None,
                 initial_key: Optional[str] = None,
                 initial_changed: Optional[bool] = None) -> str:
        if initial_key is not None:
            return initial_key

        meta = 'meta' if product.generic is None else 'range meta'
        if initial_changed:
            end = '0 ends all' if item is None else '0 ends or discards meta'
            skip = f'empty ends this {meta}, {end}, edit, view'
        else:
            skip = f'empty or 0 skips {meta}, edit'

        return self._input.get_input(f'Metadata key ({skip}, ? menu, ! cancel)',
                                     str, options='meta')

    def _get_value(self, product: Product, item: Optional[ProductItem],
                   key: str) -> Input:
        prompt = key.title()
        has_value = False
        default: Optional[Input] = None
        if key in self.MATCHERS:
            input_type = self.MATCHERS[key].get('input_type', str)
            options = self.MATCHERS[key].get('options')
            has_value = bool(getattr(product, f'{key}s'))
            if not has_value and item is not None:
                default = getattr(item, key, None)
            if default is not None and 'normalize' in self.MATCHERS[key]:
                normalize = getattr(item, self.MATCHERS[key]['normalize'])
                default = input_type(Quantity(default / normalize).amount)
        elif key in OPTIONAL_FIELDS:
            input_type = Product.__table__.c[key].type.python_type
            options = f'{key}s'
            has_value = getattr(product, key) is not None
        else:
            raise KeyError(key)

        if key == MapKey.MAP_SKU.value:
            prompt = 'Shop-specific SKU'
        elif key == MapKey.MAP_GTIN.value:
            prompt = 'GTIN-14/EAN (barcode)'
            input_type = GTIN

        if has_value:
            clear = "empty" if input_type == str else "negative"
            prompt = f'{prompt} ({clear} to clear field)'

        return self._input.get_input(prompt, input_type, options=options,
                                     default=default)

    def _set_key_value(self, product: Product, item: Optional[ProductItem],
                       key: str, value: Input) -> None:
        if isinstance(value, (Price, Quantity, int)):
            empty = value < 0
        else:
            empty = value == ""

        if key in self.MATCHERS:
            # Handle label/price/discount differently by adding to list or
            # removing if kept default with a matcher list or no item
            if empty:
                setattr(product, f'{key}s', [])
            else:
                try:
                    attrs = self._get_extra_key_value(product, item, key)
                except ValueError as e:
                    LOGGER.warning('Could not add %s: %r', key, e)
                    return

                attrs[self.MATCHERS[key]['key']] = value
                matcher = self.MATCHERS[key]['model'](**attrs)
                getattr(product, f'{key}s').append(matcher)
        else:
            setattr(product, key, value if not empty else None)

    def _get_extra_key_value(self, product: Product,
                             item: Optional[ProductItem],
                             key: str) -> dict[str, Input]:
        matcher_attrs: dict[str, Input] = {}
        if 'extra_key' in self.MATCHERS[key]:
            extra_key = self.MATCHERS[key]['extra_key']
            plain = any(price.indicator is None for price in product.prices)
            if not plain:
                if item is not None and item.unit is not None:
                    default = str(item.unit)
                else:
                    default = None
                indicator = self._input.get_input(extra_key.title(), str,
                                                  options=f'{extra_key}s',
                                                  default=default)
                if indicator != '':
                    matcher_attrs[extra_key] = indicator
                elif product.prices:
                    raise ValueError('All matchers must have indicators')

        return matcher_attrs

    def _find_duplicate(self, product: Product) -> Optional[Product]:
        existing = self._matcher.check_map(product)
        if product.generic is not None and \
            (existing is None or existing.generic == product):
            matcher = ProductMatcher(map_keys={MapKey.MAP_SKU, MapKey.MAP_GTIN})
            matcher.clear_map()
            for similar in product.generic.range:
                clash = matcher.check_map(similar)
                if clash is not None and product in {similar, clash}:
                    return similar if product == clash else clash
                matcher.add_map(similar)

        return existing

    def _check_duplicate(self, product: Product) -> _MetaResult:
        existing = self._find_duplicate(product)
        while existing is not None and existing.generic != product:
            LOGGER.warning('Product metadata existing: %r', existing)
            merge_ids = self._generate_merge_ids(existing)
            id_text = ", ".join(merge_ids)
            if existing.generic is None:
                id_text = f"{id_text} or negative to add to range"
            prompt = f'Confirm merge by ID ({id_text}), empty to discard or key'
            confirm = self._input.get_input(prompt, str, options='meta')
            if not self.CONFIRM_ID.match(confirm):
                LOGGER.debug('Not an ID, so empty or key: %r', confirm)
                return confirm != '', confirm, True

            try:
                if confirm in merge_ids:
                    self._merge(product, merge_ids[confirm])
                    return False, None, True
                if int(confirm) < 0 and existing.generic is None:
                    product.merge(self._get_initial_range(existing),
                                  override=False)
                    product.generic = existing
                    return False, None, True
                LOGGER.warning('Invalid ID: %s', confirm)
            except ValueError:
                LOGGER.exception('Could not merge product metadata')

        return True, None, True

    @staticmethod
    def _generate_merge_ids(existing: Product) -> dict[str, Product]:
        merge_ids = {
            str(existing.id if existing.id is not None else "0"): existing
        }
        merge_ids.update({
            str(index + 1 if sub.id is None else sub.id): sub
            for index, sub in enumerate(existing.range)
        })
        return merge_ids

    def _merge(self, product: Product, existing: Product) -> None:
        product.generic = None
        product.merge(existing, override=False)
        generic = existing.generic
        if generic is not None:
            generic.range[generic.range.index(existing)] = product
            product.generic_id = generic.id
        for item in self._receipt.products:
            if item.product == existing:
                item.product = product
        self._products.discard(existing)
        self._matcher.discard_map(existing)

    @property
    def description(self) -> str:
        return "Create product matching metadata"

class View(Step):
    """
    Step to display the receipt in its YAML representation.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource,
                 products: Optional[ProductsMeta] = None) -> None:
        super().__init__(receipt, input_source)
        self._products = products

    def run(self) -> ResultMeta:
        output = self._input.get_output()

        print(file=output)
        print("Prepared receipt:", file=output)
        writer = ReceiptWriter(Path(self._receipt.filename), (self._receipt,))
        writer.serialize(output)

        if self._products:
            print(file=output)
            print("Prepared product metadata:", file=output)
            products = ProductsWriter(Path("products.yml"), self._products,
                                      shared_fields=('shop',))
            products.serialize(output)

        return {}

    @property
    def description(self) -> str:
        return "View receipt in its YAML format"

class Edit(Step):
    """
    Step to edit the receipt in its YAML representation via a temporary file.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource,
                 editor: Optional[str] = None) -> None:
        super().__init__(receipt, input_source)
        self.editor = editor

    def run(self) -> ResultMeta:
        with tempfile.NamedTemporaryFile('w', suffix='.yml') as tmp_file:
            tmp_path = Path(tmp_file.name)
            writer = ReceiptWriter(tmp_path, (self._receipt,))
            writer.write()

            self.execute_editor(tmp_file.name)

            reader = ReceiptReader(tmp_path, updated=self._receipt.updated)
            try:
                receipt = next(reader.read())
                # Replace receipt
                update_path = self._receipt.date != receipt.date or \
                    self._receipt.shop != receipt.shop
                self._receipt.date = receipt.date
                self._receipt.shop = receipt.shop
                self._receipt.products = receipt.products
                self._receipt.discounts = receipt.discounts
                return {'receipt_path': update_path}
            except (StopIteration, TypeError, ValueError) as error:
                raise ReturnToMenu('Invalid or missing edited receipt YAML') \
                    from error

    def execute_editor(self, filename: str) -> None:
        """
        Open an editor to edit the provided filename.
        """

        # Find editor which can be found in the PATH
        editors = [
            self.editor, os.getenv('VISUAL'), os.getenv('EDITOR'),
            'editor', 'vim'
        ]
        for editor in editors:
            if editor is not None and \
                shutil.which(editor.split(' ', 1)[0]) is not None:
                break
        else:
            raise ReturnToMenu('No editor executable found')

        # Spawn selected editor
        try:
            subprocess.run(editor.split(' ') + [filename], check=True)
        except subprocess.CalledProcessError as exit_status:
            raise ReturnToMenu('Editor returned non-zero exit status') \
                from exit_status

    @property
    def description(self) -> str:
        return "Edit the current receipt via its YAML format"

class Write(Step):
    """
    Final step to write the receipt to a YAML file and store in the database.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource,
                 matcher: ProductMatcher, products: ProductsMeta) -> None:
        super().__init__(receipt, input_source)
        # Path should be updated based on new metadata
        self.path = Path(receipt.filename)
        self._products = products
        self._matcher = matcher

    def run(self) -> ResultMeta:
        if not self._receipt.products:
            raise ReturnToMenu('No products added to receipt')

        writer = ReceiptWriter(self.path, (self._receipt,))
        writer.write()
        with Database() as session:
            self._matcher.discounts = True
            candidates = self._matcher.find_candidates(session,
                                                       self._receipt.products,
                                                       self._products)
            pairs = self._matcher.filter_duplicate_candidates(candidates)
            for product, item in pairs:
                LOGGER.info('Matching %r to %r', item, product)
                item.product = product
            if self._products:
                inventory = ProductInventory.select(session)
                updates = ProductInventory.spread(self._products)
                LOGGER.debug('%r %r', updates, self._products)
                inventory.merge_update(updates).write()

            session.merge(self._receipt)

        return {}

    @property
    def description(self) -> str:
        if self._products:
            return "Write completed receipt and product metadata, then exit"
        return "Write the completed receipt and exit"

    @property
    def final(self) -> bool:
        return True

class Quit(Step):
    """
    Step to exit the receipt creation menu.
    """

    def run(self) -> ResultMeta:
        LOGGER.warning('Discarding entire receipt')
        return {}

    @property
    def description(self) -> str:
        return "Exit the receipt creation menu without writing"

    @property
    def final(self) -> bool:
        return True

class Help(Step):
    """
    Step to display help for steps that are usable from the menu.
    """

    def __init__(self, receipt: Receipt, input_source: InputSource):
        super().__init__(receipt, input_source)
        self.menu: Menu = {}

    @property
    def description(self) -> str:
        return "View this usage help message"

    def run(self) -> ResultMeta:
        output = self._input.get_output()
        choice_length = len(max(self.menu, key=len))
        for choice, step in self.menu.items():
            print(f"{choice: <{choice_length}} {step.description}", file=output)

        print("Initial characters match the first option with that prefix.",
              file=output)
        return {}



"""
Subcommand to create a new receipt YAML file and import it.
"""

from datetime import datetime, date, time, timedelta
import os
from pathlib import Path
import sys
from typing import Optional, Sequence, TextIO, TypeVar, Union, TYPE_CHECKING
from sqlalchemy import select
from sqlalchemy.orm import Session
from sqlalchemy.sql.functions import min as min_, max as max_
from .input import InputSource, Prompt
from .step import Menu, ProductsMeta, ResultMeta, ReturnToMenu, Step, \
    Read, Products, Discounts, ProductMeta, View, Write, Edit, Quit, Help
from ..base import Base
from ...database import Database
from ...io.products import OPTIONAL_FIELDS
from ...matcher.product import ProductMatcher
from ...models.product import Product
from ...models.receipt import Discount, ProductItem, Receipt
from ...models.shop import Shop

@Base.register("new")
class New(Base):
    """
    Create a YAML file for a receipt and import it to the database.
    """

    subparser_keywords = {
        'help': 'Create receipt file and import',
        'description': 'Interactively fill in a YAML file for a receipt and '
                       'import it to the database.'
    }
    subparser_arguments = [
        (('-c', '--confirm'), {
            'action': 'store_true',
            'default': False,
            'help': 'Confirm before updating database files or exiting'
        })
    ]

    def __init__(self) -> None:
        super().__init__()
        self.confirm = False

    def _get_menu_step(self, menu: Menu, input_source: InputSource) -> Step:
        choice: Optional[str] = None
        while choice not in menu:
            choice = input_source.get_input('Menu (help or ? for usage)', str,
                                            options='menu')
            if choice != '' and choice not in menu:
                # Autocomplete
                choice = input_source.get_completion(choice, 0)

        return menu[choice]

    def _show_menu_step(self, menu: Menu, step: Step,
                        reason: ReturnToMenu) -> Step:
        if reason.msg:
            self.logger.warning('%s', reason.msg)
        if step.final:
            step = menu['view']
            step.run()
        return step

    def _confirm_final(self, step: Step, input_source: InputSource) -> None:
        if self.confirm and step.final:
            prompt = f'Confirm that you want to {step.description.lower()} (y)'
            if input_source.get_input(prompt, str) != 'y':
                raise ReturnToMenu('Confirmation canceled')

    def _get_path(self, receipt_date: datetime, shop: str) -> Path:
        data_path = Path(self.settings.get('data', 'path'))
        data_format = self.settings.get('data', 'format')
        filename = data_format.format(date=receipt_date, shop=shop)
        return Path(data_path) / filename

    def _load_date_suggestions(self, session: Session,
                               input_source: InputSource) -> None:
        indicators = [ProductMatcher.IND_MINIMUM, ProductMatcher.IND_MAXIMUM]
        dates = session.execute(select(min_(Receipt.date).label("min"),
                                       max_(Receipt.date).label("max"))).first()
        if dates is None or dates.min is None:
            input_source.update_suggestions({'indicators': indicators})
            return

        today = date.today()
        years = range(dates.min.year, today.year + 1)
        input_source.update_suggestions({
            'days': [
                str(dates.max + timedelta(days=day))
                for day in range(max(0, (today - dates.max).days) + 1)
            ],
            'indicators': [str(year) for year in years] + indicators
        })

    def _load_suggestions(self, session: Session,
                          input_source: InputSource) -> None:
        self._load_date_suggestions(session, input_source)
        input_source.update_suggestions({
            'shops': list(session.scalars(select(Shop.key)
                                          .order_by(Shop.key))),
            'products': list(session.scalars(select(ProductItem.label)
                                             .distinct()
                                             .order_by(ProductItem.label))),
            'discounts': list(session.scalars(select(Discount.label)
                                              .distinct()
                                              .order_by(Discount.label))),
            'meta': ['label', 'price', 'discount'] + list(OPTIONAL_FIELDS) + [
                'range', 'view'
            ]
        })

    def run(self) -> None:
        input_source: InputSource = Prompt()
        matcher = ProductMatcher()
        matcher.discounts = False

        with Database() as session:
            self._load_suggestions(session, input_source)

        receipt_date = input_source.get_date(datetime.combine(date.today(),
                                                              time.min))
        shop = input_source.get_input('Shop', str, options='shops')
        path = self._get_path(receipt_date, shop)
        receipt = Receipt(filename=path.name, updated=datetime.now(),
                          date=receipt_date.date(), shop=shop)
        products: ProductsMeta = set()
        write = Write(receipt, input_source, matcher=matcher, products=products)
        write.path = path
        usage = Help(receipt, input_source)
        menu: Menu = {
            'read': Read(receipt, input_source, matcher=matcher),
            'products': Products(receipt, input_source, matcher=matcher,
                                 products=products),
            'discounts': Discounts(receipt, input_source, matcher=matcher),
            'meta': ProductMeta(receipt, input_source, matcher=matcher,
                                products=products),
            'view': View(receipt, input_source, products=products),
            'write': write,
            'edit': Edit(receipt, input_source,
                         editor=self.settings.get('data', 'editor')),
            'quit': Quit(receipt, input_source),
            'help': usage,
            '?': usage
        }
        usage.menu = menu
        step = self._run_sequential(menu, input_source)
        if step.final:
            return

        # Sequential run did not lead to a final step, so ask for menu choice
        input_source.update_suggestions({'menu': list(menu.keys())})
        while not step.final:
            step = self._get_menu_step(menu, input_source)
            try:
                self._confirm_final(step, input_source)
                result = step.run()
                # Edit might change receipt metadata
                if result.get('receipt_path', False):
                    if receipt.date != receipt_date.date():
                        receipt_date = datetime.combine(receipt.date, time.min)
                    write.path = self._get_path(receipt_date, receipt.shop)
                    receipt.filename = write.path.name
            except ReturnToMenu as reason:
                step = self._show_menu_step(menu, step, reason)

    def _run_sequential(self, menu: Menu, input_source: InputSource) -> Step:
        if not menu: # pragma: no cover
            raise ValueError('Menu must have defined steps')
        step: Step
        for step in menu.values(): # pragma: no branch
            try:
                self._confirm_final(step, input_source)
                step.run()
                if step.final:
                    return step
            except ReturnToMenu as reason:
                step = self._show_menu_step(menu, step, reason)
                break

        return step



"""
Subcommand to match entities in the database.
"""

from .base import Base
from ..database import Database
from ..matcher.product import ProductMatcher

@Base.register("match")
class Match(Base):
    """
    Update entities with references to metadata based on matching patterns.
    """

    subparser_keywords = {
        'help': 'Connect receipt product items to metadata',
        'description': 'Match products based on labels, prices and discounts '
                       'and connect the items to their product metadata.'
    }
    subparser_arguments = [
        (('-u', '--update'), {
            'action': 'store_true',
            'default': False,
            'help': 'Also update existing matched product references'
        })
    ]

    def __init__(self) -> None:
        super().__init__()
        self.update = False

    def run(self) -> None:
        with Database() as session:
            matcher = ProductMatcher()
            pairs = matcher.find_candidates(session,
                                            only_unmatched=not self.update)
            for product, item in matcher.filter_duplicate_candidates(pairs):
                self.logger.info('Matching %r with %r', item, product)
                item.product = product



"""
Subcommand to remove receipt YAML file(s) from data path and database.
"""

from pathlib import Path
from sqlalchemy import delete
from .base import Base
from ..database import Database
from ..models import Receipt

@Base.register("delete")
class Delete(Base):
    """
    Delete YAML files and database entries for receipts.
    """

    subparser_keywords = {
        'aliases': ['rm'],
        'help': 'Delete receipt files and/or database entries',
        'description': 'Delete YAML files for receipts from the data paths and '
                       'from the database.'
    }
    subparser_arguments = [
        ('files', {
            'metavar': 'FILE',
            'nargs': '+',
            'help': 'One or more files to delete'
        }),
        (('-k', '--keep'), {
            'action': 'store_true',
            'default': False,
            'help': 'Do not delete YAML file from data path'
        })
    ]

    def __init__(self) -> None:
        super().__init__()
        self.files: list[str] = []
        self.keep = False

    def run(self) -> None:
        data_path = Path(self.settings.get('data', 'path'))
        data_pattern = self.settings.get('data', 'pattern')

        # Filter off path elements to just keep the file name
        files = tuple(Path(file).name for file in self.files)

        with Database() as session:
            session.execute(delete(Receipt).where(Receipt.filename.in_(files)))

        if not self.keep:
            for file in files:
                try:
                    next(data_path.glob(f"{data_pattern}/{file}")).unlink()
                except (StopIteration, FileNotFoundError):
                    self.logger.warning("File not found in data path: %s", file)



"""
Products inventory.
"""

from collections.abc import Iterable, Iterator, Sequence
import glob
import logging
from pathlib import Path
import re
from string import Formatter
from typing import Optional
from sqlalchemy import select
from sqlalchemy.orm import Session, selectinload
from .base import Inventory, Selectors
from ..io.products import ProductsReader, ProductsWriter, SharedFields, \
    SHARED_FIELDS
from ..matcher.product import ProductMatcher
from ..models.product import Product
from ..settings import Settings

LOGGER = logging.getLogger(__name__)

class Products(dict, Inventory[Product]):
    """
    Inventory of products grouped by their identifying fields.
    """

    def __init__(self, mapping = None, /, parts: Optional[SharedFields] = None):
        super().__init__()
        if mapping is not None:
            self.update(mapping)
        if parts is None:
            settings = Settings.get_settings()
            parts = self.get_parts(settings)[2]
        self._parts = set(parts)

    @staticmethod
    def get_parts(settings: Settings) \
            -> tuple[str, str, SharedFields, re.Pattern[str]]:
        """
        Retrieve various formatting, selecting and matching parts for inventory
        filenames of products.
        """

        formatter = Formatter()
        path_format = settings.get('data', 'products')
        prefixes, keys, _, _ = zip(*formatter.parse(path_format))
        glob_pattern = '*'.join(glob.escape(prefix) for prefix in prefixes)
        fields = tuple(key for key in keys if key in SHARED_FIELDS)
        path = ''.join(rf"{re.escape(prefix)}(?P<{key}>.*)??"
                       if key is not None else re.escape(prefix)
                       for prefix, key in zip(prefixes, keys))
        pattern = re.compile(rf"(^|.*/){path}$")
        return path_format, glob_pattern, fields, pattern

    @classmethod
    def spread(cls, models: Iterable[Product]) -> "Inventory[Product]":
        inventory: dict[Path, list[Product]] = {}
        settings = Settings.get_settings()
        data_path = settings.get('data', 'path')
        path_format, _, parts, _ = cls.get_parts(settings)
        for model in models:
            fields = {str(part): getattr(model, part) for part in parts}
            path = data_path / Path(path_format.format(**fields))
            inventory.setdefault(path.resolve(), [])
            inventory[path.resolve()].append(model)

        return cls(inventory, parts=parts)

    @classmethod
    def select(cls, session: Session,
               selectors: Optional[Selectors] = None) -> "Inventory[Product]":
        inventory: dict[Path, Sequence[Product]] = {}
        settings = Settings.get_settings()
        data_path = settings.get('data', 'path')
        path_format, _, parts, _ = cls.get_parts(settings)
        if not parts:
            selectors = [{}]
        elif not selectors:
            query = select(*(getattr(Product, field) for field in parts)) \
                .distinct()
            selectors = [
                dict(zip(parts, values)) for values in session.execute(query)
            ]
            LOGGER.debug('Products files fields: %r', selectors)

        for fields in selectors:
            products = session.scalars(select(Product)
                                       .options(selectinload(Product.range),
                                                selectinload(Product.generic))
                                       .filter(Product.generic_id.is_(None))
                                       .filter_by(**fields)).all()
            path = data_path / Path(path_format.format(**fields))
            inventory[path.resolve()] = products

        return cls(inventory, parts=parts)

    @classmethod
    def read(cls) -> "Inventory[Product]":
        inventory: dict[Path, Sequence[Product]] = {}
        settings = Settings.get_settings()
        data_path = Path(settings.get('data', 'path'))
        _, glob_pattern, parts, _ = cls.get_parts(settings)
        for path in sorted(data_path.glob(glob_pattern)):
            LOGGER.info('Looking at products in %s', path)
            try:
                products = list(ProductsReader(path).read())
                inventory[path.resolve()] = products
            except (TypeError, ValueError):
                LOGGER.exception('Could not parse product from %s', path)

        return cls(inventory, parts=parts)

    def get_writers(self) -> Iterator[ProductsWriter]:
        for path, products in self.items():
            yield ProductsWriter(path, products, shared_fields=self._parts)

    def _find_match(self, product: Product, matcher: ProductMatcher,
                    update: bool = True,
                    only_new: bool = False) -> tuple[Optional[Product], bool]:
        changed = False
        existing = matcher.check_map(product)
        if existing is None:
            changed = True
            existing = product
        elif only_new:
            return None, False
        elif not update:
            existing = existing.copy()
        if existing.merge(product):
            changed = True
        return existing, changed

    def merge_update(self, other: "Inventory[Product]", update: bool = True,
                     only_new: bool = False) -> "Inventory[Product]":
        updated: dict[Path, list[Product]] = {}
        matcher = ProductMatcher()
        matcher.fill_map(self)
        if only_new:
            update = False
        for path, products in other.items():
            changed = False
            updates: list[Product] = []
            for product in products:
                match, change = self._find_match(product, matcher,
                                                 update=update,
                                                 only_new=only_new)
                changed = changed or change
                if match is not None:
                    updates.append(match)

            if update:
                self.setdefault(path, [])
                self[path].extend(change for change in updates
                                  if change not in self[path])
                # Make the updates follow the same order and have entire path
                updates = self[path].copy()
            if changed:
                updated[path] = updates

        return Products(updated, parts=self._parts)



"""
Add product matchers

Revision ID: aaff17ed83d8
Revises: d6854aeb9756
Create Date: 2025-04-25 18:08:04.816341
"""
# pylint: disable=invalid-name

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision: str = 'aaff17ed83d8'
down_revision: Union[str, None] = 'd6854aeb9756'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """
    Perform the upgrade.
    """

    product_ref = 'product.id'

    # ### commands auto generated by Alembic - please adjust! ###
    discount_fk = 'fk_product_discount_match_product_id_product'
    discount_pk = 'pk_product_discount_match'
    op.create_table('product_discount_match',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('product_id', sa.Integer(), nullable=False),
                    sa.Column('label', sa.String(), nullable=False),
                    sa.ForeignKeyConstraint(['product_id'], [product_ref],
                                            name=op.f(discount_fk),
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id', name=op.f(discount_pk)))

    label_fk = 'fk_product_label_match_product_id_product'
    label_pk = 'pk_product_label_match'
    op.create_table('product_label_match',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('product_id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.String(), nullable=False),
                    sa.ForeignKeyConstraint(['product_id'], [product_ref],
                                            name=op.f(label_fk),
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id', name=op.f(label_pk)))

    price_fk = 'fk_product_price_match_product_id_product'
    price_pk = 'pk_product_price_match'
    op.create_table('product_price_match',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('product_id', sa.Integer(), nullable=False),
                    sa.Column('value', sa.Numeric(scale=2), nullable=False),
                    sa.Column('indicator', sa.String(), nullable=True),
                    sa.ForeignKeyConstraint(['product_id'], [product_ref],
                                            name=op.f(price_fk),
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id', name=op.f(price_pk)))
    # ### end Alembic commands ###


def downgrade() -> None:
    """
    Perform the downgrade.
    """

    op.drop_table('product_price_match')
    op.drop_table('product_label_match')
    op.drop_table('product_discount_match')
    # ### end Alembic commands ###



"""
Products matching metadata file handling.
"""

from datetime import datetime
from pathlib import Path
from typing import get_args, Collection, Iterable, Iterator, IO, Literal, \
    Optional, TypeVar, Union
from typing_extensions import TypedDict
from .base import YAMLReader, YAMLWriter
from ..models.base import GTIN, Price, Quantity
from ..models.product import Product, LabelMatch, PriceMatch, DiscountMatch

class _Product(TypedDict, total=False):
    """
    Serialized product metadata.
    """

    shop: str
    labels: list[str]
    prices: Union[list[Price], dict[str, Price]]
    bonuses: list[str]
    brand: Optional[str]
    description: Optional[str]
    category: Optional[str]
    type: Optional[str]
    portions: Optional[int]
    weight: Optional[Quantity]
    volume: Optional[Quantity]
    alcohol: Optional[str]
    sku: str
    gtin: int

class _GenericProduct(_Product, total=False):
    range: list["_Product"]

class _InventoryGroup(TypedDict, total=False):
    shop: str
    brand: str
    category: str
    type: str
    products: list[_GenericProduct]

PrimaryField = Literal["shop"]
OptionalShareableField = Literal["brand", "category", "type"]
ShareableField = Literal[PrimaryField, OptionalShareableField]
SharedFields = Iterable[ShareableField]
PropertyField = Literal[
    OptionalShareableField,
    "description", "portions", "weight", "volume", "alcohol"
]
IdentifierField = Literal["sku", "gtin"]
Field = Literal[PrimaryField, PropertyField, IdentifierField]
OptionalField = Literal[PropertyField, IdentifierField]
_Input = Union[str, int, Quantity]
_FieldT = TypeVar("_FieldT", bound=_Input)
SHARED_FIELDS: tuple[ShareableField, ...] = get_args(ShareableField)
PROPERTY_FIELDS: tuple[PropertyField, ...] = get_args(PropertyField)
IDENTIFIER_FIELDS: tuple[IdentifierField, ...] = get_args(IdentifierField)
OPTIONAL_FIELDS: tuple[OptionalField, ...] = get_args(OptionalField)

class ProductsReader(YAMLReader[Product]):
    """
    File reader for products metadata.
    """

    def parse(self, file: IO) -> Iterator[Product]:
        data: _InventoryGroup = self.load(file)
        if not isinstance(data, dict):
            raise TypeError(f"File '{self._path}' does not contain a mapping")
        if not isinstance(data.get('products'), list):
            raise TypeError(f"File '{self._path}' is missing 'products' list")

        for meta in data['products']:
            product = self._product(data, {}, meta)
            product.range = [
                self._product(data, meta, sub_meta)
                for sub_meta in meta.get('range', [])
            ]
            yield product

    @staticmethod
    def _get(input_type: type[_FieldT],
             value: Optional[_Input]) -> Optional[_FieldT]:
        if value is not None:
            value = input_type(value)
            if not isinstance(value, input_type): # pragma: no cover
                value = None

        return value

    def _product(self, data: _InventoryGroup, generic: _GenericProduct,
                 meta: _Product) -> Product:
        if not isinstance(meta, dict):
            raise TypeError(f"Product is not a mapping: {meta!r}")
        product = Product(shop=data.get('shop', generic.get('shop')),
                          brand=meta.get('brand', generic.get('brand')),
                          description=meta.get('description',
                                               generic.get('description')),
                          category=meta.get('category',
                                            generic.get('category',
                                                        data.get('category'))),
                          type=meta.get('type', generic.get('type',
                                                            data.get('type'))),
                          portions=self._get(int,
                                             meta.get('portions',
                                                      generic.get('portions'))),
                          weight=self._get(Quantity,
                                           meta.get('weight',
                                                    generic.get('weight'))),
                          volume=self._get(Quantity,
                                           meta.get('volume',
                                                    generic.get('volume'))),
                          alcohol=meta.get('alcohol', generic.get('alcohol')),
                          sku=meta.get('sku'),
                          gtin=GTIN(meta['gtin']) if 'gtin' in meta else None)

        product.labels = [
            LabelMatch(name=name)
            for name in meta.get('labels', generic.get('labels', []))
        ]
        prices = meta.get('prices', generic.get('prices', []))
        if isinstance(prices, list):
            product.prices = [
                PriceMatch(value=Price(price)) for price in prices
            ]
        else:
            product.prices = [
                PriceMatch(value=Price(price), indicator=key)
                for key, price in prices.items()
            ]
        product.discounts = [
            DiscountMatch(label=label)
            for label in meta.get('bonuses', generic.get('bonuses', []))
        ]

        return product

class ProductsWriter(YAMLWriter[Product]):
    """
    File writer for products metadata.
    """

    def __init__(self, path: Path, models: Collection[Product],
                 updated: Optional[datetime] = None,
                 shared_fields: SharedFields = ('shop', 'category', 'type')):
        super().__init__(path, models, updated=updated)
        self._shared_fields = set(shared_fields)

    @staticmethod
    def _get_prices(product: Product) -> Union[list[Price], dict[str, Price]]:
        prices: list[Price] = []
        indicator_prices: dict[str, Price] = {}

        for price in product.prices:
            if price.indicator is not None:
                indicator_prices[price.indicator] = price.value
            else:
                prices.append(price.value)

        if indicator_prices:
            if prices:
                raise ValueError('Not all price matchers have indicators')
            return indicator_prices

        return prices

    def _get_product(self, product: Product, skip_fields: set[Field],
                     generic: _GenericProduct) \
            -> Union[_Product, _GenericProduct]:
        data: Union[_Product, _GenericProduct] = {}
        if 'shop' not in skip_fields:
            data['shop'] = product.shop

        labels = [label.name for label in product.labels]
        if labels != generic.get('labels', []):
            data['labels'] = labels

        prices = self._get_prices(product)
        if prices != generic.get('prices', []):
            data['prices'] = prices

        discounts = [discount.label for discount in product.discounts]
        if discounts != generic.get('bonuses', []):
            data['bonuses'] = discounts

        for field in PROPERTY_FIELDS:
            if field not in skip_fields:
                value = getattr(product, field, None)
                if value != generic.get(field):
                    data[field] = value
        for id_field in IDENTIFIER_FIELDS:
            identifier = getattr(product, id_field, None)
            if identifier is not None:
                data[id_field] = identifier

        return data

    def _get_generic_product(self, product: Product, skip_fields: set[Field]) \
            -> _GenericProduct:
        data: _GenericProduct = {**self._get_product(product, skip_fields, {})}

        if product.range:
            data['range'] = [
                self._get_product(sub_product, skip_fields | {'shop'}, data)
                for sub_product in product.range
            ]

        return data

    def serialize(self, file: IO) -> None:
        group: _InventoryGroup = {}
        skip_fields: set[Field] = set()
        for shared in self._shared_fields:
            values = {getattr(product, shared) for product in self._models}
            try:
                common = values.pop()
            except KeyError:
                common = None
            if not values and common is not None:
                group[shared] = str(common)
                skip_fields.add(shared)
            elif shared == 'shop':
                raise ValueError('Not all products are from the same shop')

        group['products'] = [
            self._get_generic_product(product, skip_fields)
            for product in self._models
        ]
        self.save(group, file)



"""
Subcommand collection package.
"""

from .base import Base
from .alembic import Alembic
from .config import Config
from .create import Create
from .delete import Delete
from .dump import Dump
from .match import Match
from .new import New
from .read import Read

__all__ = ["Base"]



"""
Models for product metadata.
"""

from itertools import zip_longest
import logging
from typing import Optional
from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import MappedColumn, Relationship, mapped_column, \
    relationship
from .base import Base, GTIN, Price, Quantity

LOGGER = logging.getLogger(__name__)

_CASCADE_OPTIONS = "all, delete-orphan"
_PRODUCT_REF = "product.id"

class Product(Base): # pylint: disable=too-few-public-methods
    """
    Product model for metadata.
    """

    __tablename__ = "product"

    id: MappedColumn[int] = mapped_column(primary_key=True, autoincrement=True)
    shop: MappedColumn[str] = mapped_column(String(32)) # shop.key

    # Matchers
    labels: Relationship[list["LabelMatch"]] = \
        relationship(back_populates="product", cascade=_CASCADE_OPTIONS,
                     passive_deletes=True, lazy="selectin")
    prices: Relationship[list["PriceMatch"]] = \
        relationship(back_populates="product", cascade=_CASCADE_OPTIONS,
                     passive_deletes=True, lazy="selectin")
    discounts: Relationship[list["DiscountMatch"]] = \
        relationship(back_populates="product", cascade=_CASCADE_OPTIONS,
                     passive_deletes=True, lazy="selectin")

    # Descriptors
    brand: MappedColumn[Optional[str]]
    description: MappedColumn[Optional[str]]

    # Taxonomy
    category: MappedColumn[Optional[str]]
    type: MappedColumn[Optional[str]]

    # Trade item properties
    portions: MappedColumn[Optional[int]]
    weight: MappedColumn[Optional[Quantity]]
    volume: MappedColumn[Optional[Quantity]]
    alcohol: MappedColumn[Optional[str]]

    # Shop-specific and globally unique identifiers
    sku: MappedColumn[Optional[str]]
    gtin: MappedColumn[Optional[GTIN]]

    # Product range differentiation
    range: Relationship[list["Product"]] = \
        relationship(back_populates="generic", cascade=_CASCADE_OPTIONS,
                     passive_deletes=True, order_by="Product.id",
                     lazy="selectin")
    generic_id: MappedColumn[Optional[int]] = \
        mapped_column(ForeignKey(_PRODUCT_REF, ondelete="CASCADE"))
    generic: Relationship[Optional["Product"]] = \
        relationship(back_populates="range", remote_side=[id], lazy="selectin")

    def clear(self) -> None:
        """
        Remove all matchers, properties, identifiers and range products, but
        not the generic product or its properties that we inherit now.
        """

        self.labels = []
        self.prices = []
        self.discounts = []
        self.range = []
        for column, meta in self.__table__.c.items():
            if meta.nullable and not meta.foreign_keys:
                setattr(self, column, None)

        # Obtain inherited default properties
        if self.generic is not None:
            self.merge(self.generic)
            self.sku = None
            self.gtin = None

    def replace(self, new: "Product") -> None:
        """
        Replace all matchers, properties, identifiers and range products with
        those defined in the `new` product, or with the generic product's
        inherited properties; the original generic product is kept.
        """

        self.clear()

        # Clear matchers obtained from generic product in favor of overrides
        if self.generic is not None:
            if new.labels:
                self.labels = []
            if new.prices:
                self.prices = []
            if new.discounts:
                self.discounts = []

        self.merge(new)

    def copy(self) -> "Product":
        """
        Copy the product.
        """

        copy = Product(shop=self.shop)
        copy.merge(self)
        return copy

    def check_merge(self, other: "Product") -> None:
        """
        Check if the other product is compatible with merging into this product.
        """

        if self.prices and other.prices:
            plain = any(price.indicator is None for price in self.prices)
            other_plain = any(price.indicator is None for price in other.prices)
            if plain ^ other_plain:
                raise ValueError("Both products' price matchers must have "
                                 "indicators, or none of theirs should: "
                                 f"{self!r} {other!r}")
        for product_range, other_range in zip(self.range, other.range):
            product_range.check_merge(other_range)

    def _merge_range(self, other: "Product", override: bool = True) -> bool:
        changed = False
        if self.generic is None:
            for sub_range, other_range in zip_longest(self.range, other.range):
                if sub_range is None:
                    LOGGER.debug('Adding range product %r', other_range)
                    self.range.append(other_range.copy())
                    changed = True
                elif other_range is not None and \
                    sub_range.merge(other_range, override=override):
                    LOGGER.debug('Merged range products')
                    changed = True

        return changed

    def _merge_fields(self, other: "Product", override: bool = True) -> bool:
        changed = False
        for column, meta in self.__table__.c.items():
            current = getattr(self, column)
            if meta.foreign_keys or (current is not None and not override):
                LOGGER.debug('Not updating field %s (%r)', column, current)
                continue

            target = getattr(other, column)
            if (meta.nullable or (meta.primary_key and current is None)) and \
                target is not None and current != target:
                LOGGER.debug('Updating field %s from %r to %r', column,
                              current, target)
                setattr(self, column, target)
                changed = True

        return changed

    def merge(self, other: "Product", override: bool = True) -> bool:
        """
        Merge attributes of the other product into this one.

        This replaces values and the primary key in this product, except for the
        shop identifier (which is always kept) and the matchers (where unique
        matchers from the other product are added).

        This is similar to a session merge except no database changes are done
        and the matchers are more deeply merged.

        If `override` is disabled, then simple property fields that already have
        a value are not changed. Matchers are always updated.

        Returns whether the product has changed, with new matchers or different
        values.
        """

        self.check_merge(other)

        LOGGER.debug('Performing merge into %r from %r', self, other)
        changed = False
        labels = {label.name for label in self.labels}
        for label in other.labels:
            if label.name not in labels:
                LOGGER.debug('Adding label matcher %s', label.name)
                self.labels.append(LabelMatch(name=label.name))
                changed = True
        prices = {(price.indicator, price.value) for price in self.prices}
        for price in other.prices:
            if (price.indicator, price.value) not in prices:
                LOGGER.debug('Adding price matcher %r (indicator: %r)',
                             price.value, price.indicator)
                self.prices.append(PriceMatch(indicator=price.indicator,
                                              value=price.value))
                changed = True
        discounts = {discount.label for discount in self.discounts}
        for discount in other.discounts:
            if discount.label not in discounts:
                LOGGER.debug('Adding discount matcher %r', discount.label)
                self.discounts.append(DiscountMatch(label=discount.label))
                changed = True

        if self._merge_range(other, override=override):
            changed = True

        if self._merge_fields(other, override=override):
            changed = True

        LOGGER.debug('Merged products: %r', changed)
        return changed

    def __repr__(self) -> str:
        weight = str(self.weight) if self.weight is not None else None
        volume = str(self.volume) if self.volume is not None else None
        sub_range = f", range={self.range!r}" if self.generic is None else ""
        return (f"Product(id={self.id!r}, shop={self.shop!r}, "
                f"labels={self.labels!r}, prices={self.prices!r}, "
                f"discounts={self.discounts!r}, brand={self.brand!r}, "
                f"description={self.description!r}, "
                f"category={self.category!r}, type={self.type!r}, "
                f"portions={self.portions!r}, weight={weight!r}, "
                f"volume={volume!r}, alcohol={self.alcohol!r}, "
                f"sku={self.sku!r}, gtin={self.gtin!r}{sub_range})")

class LabelMatch(Base): # pylint: disable=too-few-public-methods
    """
    Label model for a product matching string.
    """

    __tablename__ = "product_label_match"

    id: MappedColumn[int] = mapped_column(primary_key=True)
    product_id: MappedColumn[int] = \
        mapped_column(ForeignKey(_PRODUCT_REF, ondelete='CASCADE'))
    product: Relationship[Product] = relationship(back_populates="labels")
    name: MappedColumn[str]

    def __repr__(self) -> str:
        return repr(self.name)

class PriceMatch(Base): # pylint: disable=too-few-public-methods
    """
    Price model for a product matching value, which may be part of a value range
    or time interval.
    """

    __tablename__ = "product_price_match"

    id: MappedColumn[int] = mapped_column(primary_key=True)
    product_id: MappedColumn[int] = \
        mapped_column(ForeignKey(_PRODUCT_REF, ondelete='CASCADE'))
    product: Relationship[Product] = relationship(back_populates="prices")
    value: MappedColumn[Price]
    indicator: MappedColumn[Optional[str]]

    def __repr__(self) -> str:
        return str(self.value) if self.indicator is None else \
            f"({self.indicator!r}, {self.value!s})"

class DiscountMatch(Base): # pylint: disable=too-few-public-methods
    """
    Discount label model for a product matching string.
    """

    __tablename__ = "product_discount_match"

    id: MappedColumn[int] = mapped_column(primary_key=True)
    product_id: MappedColumn[int] = \
        mapped_column(ForeignKey(_PRODUCT_REF, ondelete='CASCADE'))
    product: Relationship[Product] = relationship(back_populates="discounts")
    label: MappedColumn[str]

    def __repr__(self) -> str:
        return repr(self.label)



"""
Attribute types for numeric values with discrete precision.
"""

from decimal import Decimal
from typing import Union
from sqlalchemy import BigInteger, Numeric
from .decorator import SerializableType

PriceNew = Union[Decimal, float, str]

class GTIN(int):
    """
    Global trade item number identifier for products.
    """

class Price(Decimal): # pylint: disable=too-few-public-methods
    """
    Price type with scale of 2 (number of decimal places).
    """

    _quantize = Decimal('1.00')

    def __new__(cls, value: PriceNew) -> "Price":
        try:
            return super().__new__(cls, Decimal(value).quantize(cls._quantize))
        except ArithmeticError as e:
            raise ValueError("Could not construct a two-decimal price") from e

class GTINType(SerializableType[GTIN, int]):
    # pylint: disable=too-many-ancestors
    """
    Type decoration handler for GTINs.
    """

    cache_ok = True
    impl = BigInteger()

    @property
    def serializable_type(self) -> type[GTIN]:
        return GTIN

    @property
    def serialized_type(self) -> type[int]:
        return int

class PriceType(SerializableType[Price, Decimal]):
    # pylint: disable=too-many-ancestors
    """
    Type decoration handler for prices.
    """

    cache_ok = True
    impl = Numeric(None, 2)

    @property
    def serializable_type(self) -> type[Price]:
        return Price

    @property
    def serialized_type(self) -> type[Decimal]:
        return Decimal



"""
Add cascade deletes for products/discounts

Revision ID: 24c54f418b92
Revises: 
Create Date: 2025-02-07 23:15:46.360221
"""
# pylint: disable=invalid-name

from typing import Sequence, Union

from alembic import op

# Revision identifiers, used by Alembic.
revision: str = '24c54f418b92'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

def upgrade() -> None:
    """
    Perform the upgrade.
    """

    with op.batch_alter_table('receipt_discount', schema=None) as batch_op:
        receipt_discount_key = 'fk_receipt_discount_receipt_key_receipt'
        batch_op.drop_constraint(receipt_discount_key, type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f(receipt_discount_key), 'receipt',
                                    ['receipt_key'], ['filename'],
                                    ondelete='CASCADE')

    with op.batch_alter_table('receipt_discount_products',
                              schema=None) as batch_op:
        discount = 'fk_receipt_discount_products_discount_id_receipt_discount'
        product = 'fk_receipt_discount_products_product_id_receipt_product'
        batch_op.drop_constraint(discount, type_='foreignkey')
        batch_op.drop_constraint(product, type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f(discount), 'receipt_discount',
                                    ['discount_id'], ['id'], ondelete='CASCADE')
        batch_op.create_foreign_key(batch_op.f(product), 'receipt_product',
                                    ['product_id'], ['id'], ondelete='CASCADE')

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        receipt_product_key = 'fk_receipt_product_receipt_key_receipt'
        batch_op.drop_constraint(receipt_product_key, type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f(receipt_product_key), 'receipt',
                                    ['receipt_key'], ['filename'],
                                    ondelete='CASCADE')

    # ### end Alembic commands ###


def downgrade() -> None:
    """
    Perform the downgrade.
    """

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        receipt_product_key = 'fk_receipt_product_receipt_key_receipt'
        batch_op.drop_constraint(batch_op.f(receipt_product_key),
                                 type_='foreignkey')
        batch_op.create_foreign_key(receipt_product_key, 'receipt',
                                    ['receipt_key'], ['filename'])

    with op.batch_alter_table('receipt_discount_products',
                              schema=None) as batch_op:
        product = 'fk_receipt_discount_products_product_id_receipt_product'
        discount = 'fk_receipt_discount_products_discount_id_receipt_discount'
        batch_op.drop_constraint(batch_op.f(product), type_='foreignkey')
        batch_op.drop_constraint(batch_op.f(discount), type_='foreignkey')
        batch_op.create_foreign_key(product, 'receipt_product',
                                    ['product_id'], ['id'])
        batch_op.create_foreign_key(discount, 'receipt_discount',
                                    ['discount_id'], ['id'])

    with op.batch_alter_table('receipt_discount', schema=None) as batch_op:
        receipt_discount_key = 'fk_receipt_discount_receipt_key_receipt'
        batch_op.drop_constraint(batch_op.f(receipt_discount_key),
                                 type_='foreignkey')
        batch_op.create_foreign_key(receipt_discount_key, 'receipt',
                                    ['receipt_key'], ['filename'])

    # ### end Alembic commands ###



"""
Increase GTIN size

Revision ID: 3b0cfa853967
Revises: aaff17ed83d8
Create Date: 2025-04-26 21:27:07.365156
"""
# pylint: disable=invalid-name

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import column, table


# Revision identifiers, used by Alembic.
revision: str = '3b0cfa853967'
down_revision: Union[str, None] = 'aaff17ed83d8'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """
    Perform the upgrade.
    """

    with op.batch_alter_table('product', schema=None) as batch_op:
        batch_op.alter_column('gtin',
                              existing_type=sa.INTEGER(),
                              type_=sa.BigInteger(),
                              existing_nullable=True)

    # ### end Alembic commands ###


def downgrade() -> None:
    """
    Perform the downgrade.
    """

    with op.batch_alter_table('product', schema=None) as batch_op:
        product = table('product', column('id', sa.Integer()),
                        column('gtin', sa.BigInteger()))
        batch_op.execute(product.update().values(gtin=None))
        batch_op.alter_column('gtin',
                              existing_type=sa.BigInteger(),
                              type_=sa.INTEGER(),
                              existing_nullable=True)

    # ### end Alembic commands ###



"""
Subcommand to generate an amalgamate settings file.
"""

import tomlkit
from tomlkit.items import Item, Table
from .base import Base
from ..settings import Settings

@Base.register("config")
class Config(Base):
    """
    Obtain settings file representation.
    """

    subparser_keywords = {
        'help': 'Obtain settings representation',
        'description': 'Generate settings TOML representation with comments.'
    }
    subparser_arguments = [
        (('section',), {
            'metavar': 'SECTION',
            'nargs': '?',
            'help': 'Optional table section name to filter on'
        }),
        (('key',), {
            'metavar': 'KEY',
            'nargs': '?',
            'help': 'Optional settings key to filter on'
        }),
        (('-f', '--file'), {
            'help': 'Generate based on specific TOML file'
        }),
        (('-p', '--prefix'), {
            'nargs': '+',
            'default': (),
            'help': 'Section prefixes in specific TOML file to look up'
        })
    ]

    def __init__(self) -> None:
        super().__init__()
        self.section: str = ''
        self.key: str = ''
        self.file: str = ''
        self.prefix: tuple[str, ...] = ()

    def run(self) -> None:
        if self.file:
            document = Settings(path=self.file, environment=False,
                                prefix=self.prefix).get_document()
        else:
            document = self.settings.get_document()

        if self.section:
            table = document.get(self.section)
            container = tomlkit.table()
            if isinstance(table, Table):
                table.trivia.indent = ''
                if self.key:
                    if self.key not in table:
                        print()
                        return

                    table = self._wrap_setting(table[self.key], self.key)

                container[self.section] = table

            print(container.as_string())
        else:
            print(document.as_string())

    def _wrap_setting(self, item: Item, key: str) -> Table:
        comments = self.settings.get_comments()
        table = tomlkit.table()
        for comment in comments.get(self.section, {}).get(key, []):
            table.add(tomlkit.comment(comment))
        table[key] = item
        return table



"""
Input source for new subcommand.
"""

from datetime import datetime
import logging
import sys
from typing import Optional, Sequence, TextIO, TypeVar, Union, TYPE_CHECKING
try:
    import readline
except ImportError:
    if not TYPE_CHECKING:
        readline = None
import dateutil.parser
from ...models.base import Price, Quantity

Input = Union[str, int, float, Price, Quantity]
InputT = TypeVar('InputT', bound=Input)

LOGGER = logging.getLogger(__name__)

class InputSource:
    """
    Abstract base class for a typed input source.
    """

    def get_input(self, name: str, input_type: type[InputT],
                  options: Optional[str] = None,
                  default: Optional[InputT] = None) -> InputT:
        """
        Retrieve an input cast to a certain type (string, integer or float).
        Optionally, the input source provides suggestions from a predefined
        completion source defined by the `options` name and may fall back to
        a `default` if nothing is input.
        """

        raise NotImplementedError('Input must be retrieved by subclasses')

    def get_date(self, default: Optional[datetime] = None) -> datetime:
        """
        Retrieve a date input. The `default` may be used as a fallback if
        nothing is input or if a partial timestamp is provided.
        """

        raise NotImplementedError('Date input be retrieved by subclasses')

    def get_output(self) -> TextIO:
        """
        Retrieve an output stream to write content to.
        """

        raise NotImplementedError('Output must be implemented by subclasses')

    def update_suggestions(self, suggestions: dict[str, list[str]]) -> None:
        """
        Include additional suggestion completion sources.
        """

    def get_completion(self, text: str, state: int) -> Optional[str]:
        """
        Retrieve a completion option for the current suggestions and text state.
        The `text` is a partial input that matches some part of the suggestions
        and `state` indicates the position of the suggestion in the sorted
        list of matching suggestions to choose.

        If there is no match found or if the input source does not support
        completion suggestions, then `None` is returned.
        """

        raise NotImplementedError('Should be implemented by subclasses')

class Prompt(InputSource):
    """
    Standard input prompt.
    """

    def __init__(self) -> None:
        self._suggestions: dict[str, list[str]] = {}
        self._options: list[str] = []
        self._matches: list[str] = []
        self.register_readline()

    def get_input(self, name: str, input_type: type[InputT],
                  options: Optional[str] = None,
                  default: Optional[InputT] = None) -> InputT:
        """
        Retrieve an input cast to a certain type (string, integer or float).
        """

        if options is not None and options in self._suggestions:
            self._options = self._suggestions[options]
        else:
            self._options = []
        value: Optional[Input] = None
        if default is not None:
            name = f'{name} (empty for "{default!s}")'
        while not isinstance(value, input_type):
            try:
                LOGGER.debug('[prompt] (%s) %s:', input_type.__name__, name)
                text = input(f'{name}: ')
                if default is not None and text == '':
                    value = default
                else:
                    value = input_type(text)
                LOGGER.debug('[prompt] input: %r', value)
            except ValueError as e:
                LOGGER.warning('Invalid %s: %s', input_type.__name__, e)
        return value

    def get_date(self, default: Optional[datetime] = None) -> datetime:
        value: Optional[datetime] = None
        if default is not None:
            day = default.isoformat(sep=' ')
        else:
            day = None
        while not isinstance(value, datetime):
            try:
                value = dateutil.parser.parse(self.get_input('Date/time', str,
                                                             options='days',
                                                             default=day),
                                              default=default)
            except ValueError as e:
                LOGGER.warning('Invalid timestamp: %s', e)
        return value

    def get_output(self) -> TextIO:
        return sys.stdout

    def update_suggestions(self, suggestions: dict[str, list[str]]) -> None:
        self._suggestions.update(suggestions)

    def get_completion(self, text: str, state: int) -> Optional[str]:
        if state == 0:
            if text == '':
                self._matches = self._options
            else:
                self._matches = [
                    option for option in self._options
                    if option.startswith(text)
                ]
        try:
            return self._matches[state]
        except IndexError:
            return None

    def display_matches(self, substitution: str, matches: Sequence[str],
                        longest_match_length: int) -> None:
        """
        Write a display of matches to the standard output compatible with
        readline buffers.
        """

        line_buffer = readline.get_line_buffer()
        output = self.get_output()
        print(file=output)

        length = int(max(map(len, matches), default=longest_match_length) * 1.2)
        template = f"{{:<{length}}}"
        buffer = ""
        for match in matches:
            display = template.format(match[len(substitution):])
            if buffer != "" and len(buffer + display) > 80:
                print(buffer, file=output)
                buffer = ""
            buffer += display

        if buffer:
            print(buffer, file=output)

        print("> ", end="", file=output)
        print(line_buffer, end="", file=output)
        output.flush()

    def register_readline(self) -> None:
        """
        Register completion method to the `readline` module.
        """

        if readline is not None: # pragma: no cover
            readline.set_completer_delims('\t\n;')
            readline.set_completer(self.get_completion)
            readline.set_completion_display_matches_hook(self.display_matches)
            readline.parse_and_bind('tab: complete')
            readline.parse_and_bind('bind ^I rl_complete')



"""
Settings module.
"""

import os
from pathlib import Path
from typing import Union
import tomlkit
from tomlkit.items import Comment, Table
from typing_extensions import Required, TypedDict

class _SettingsFile(TypedDict, total=False):
    path: Required[Union[str, os.PathLike[str]]]
    environment: bool
    prefix: tuple[str, ...]

Chain = tuple[_SettingsFile, ...]

SETTINGS_FILE_NAME = 'settings.toml'

class Settings:
    """
    Settings reader and provider.
    """

    FILES: Chain = (
        {
            'path': SETTINGS_FILE_NAME
        },
        {
            'path': 'pyproject.toml',
            'environment': False,
            'prefix': ('tool', 'rechu')
        },
        {
            'path': Path(__file__).parent / SETTINGS_FILE_NAME,
            'environment': False
        }
    )
    _files: dict[int, "Settings"] = {}

    @classmethod
    def get_settings(cls) -> "Settings":
        """
        Retrieve the settings singleton.
        """

        return cls._get_fallback(cls.FILES)

    @classmethod
    def _get_fallback(cls, fallbacks: Chain) -> "Settings":
        key = hash(tuple(tuple(file.values()) for file in fallbacks))
        if key not in cls._files:
            cls._files[key] = cls(fallbacks=fallbacks[1:], **fallbacks[0])

        return cls._files[key]

    @classmethod
    def clear(cls) -> None:
        """
        Remove the singleton instance and any fallback instances.
        """

        cls._files = {}

    def __init__(self, path: Union[str, os.PathLike[str]] = SETTINGS_FILE_NAME,
                 environment: bool = True, prefix: tuple[str, ...] = (),
                 fallbacks: Chain = ()) -> None:
        if environment:
            path = os.getenv('RECHU_SETTINGS_FILE', path)

        try:
            with Path(path).open('r', encoding='utf-8') as settings_file:
                sections = tomlkit.load(settings_file)
        except FileNotFoundError:
            sections = tomlkit.document()

        for group in prefix:
            sections = sections.get(group, {})
        self.sections: Union[Table, tomlkit.TOMLDocument] = sections

        self.environment = environment
        self.fallbacks = fallbacks
        self.prefix = prefix

    def get(self, section: str, key: str) -> str:
        """
        Retrieve a settings value from the file based on its `section` name,
        which refers to a TOML table grouping multiple settings, and its `key`,
        potentially with an environment variable override.
        """

        env_name = f"RECHU_{section.upper()}_{key.upper().replace('-', '_')}"
        if self.environment and env_name in os.environ:
            return os.environ[env_name]
        group = self.sections.get(section)
        if not isinstance(group, dict) or key not in group:
            if self.fallbacks:
                return self._get_fallback(self.fallbacks).get(section, key)
            raise KeyError(f'{section} is not a section or does not have {key}')
        return str(group[key])

    def get_comments(self) -> dict[str, dict[str, list[str]]]:
        """
        Retrieve comments of the settings by section.

        This retrieves comments for a setting from the settings file latest in
        the chain that has comments. Only comments preceding the setting are
        preserved.
        """

        comment: list[str] = []
        comments: dict[str, dict[str, list[str]]] = {}
        if self.fallbacks:
            comments = self._get_fallback(self.fallbacks).get_comments()
        for table, section in self.sections.items():
            comments.setdefault(table, {})
            for key, value in section.value.body:
                if isinstance(value, Comment):
                    comment.append(str(value).lstrip('# '))
                else:
                    # Only keep default comments
                    if key is not None and key.key not in comments[table]:
                        comments[table][key.key] = comment
                    comment = []

        return comments

    def get_document(self) -> tomlkit.TOMLDocument:
        """
        Reconstruct a TOML document with overrides from environment variables,
        default values and comments from fallbacks.
        """

        if self.fallbacks:
            document = self._get_fallback(self.fallbacks).get_document()
        else:
            document = tomlkit.document()

        comments = self.get_comments()
        for section, table in self.sections.items():
            table_comments = comments.get(section, {})
            target: Table = document.setdefault(section, tomlkit.table())
            for key in table:
                if key not in target:
                    for comment in table_comments.get(key, []):
                        target.add(tomlkit.comment(comment))
                target[key] = self.get(section, key)

        return document



"""
This is a Python script that is run whenever the alembic migration tool is
invoked. It contains instructions to configure and generate a SQLAlchemy
engine, procure a connection from that engine along with a transaction, and
then invoke the migration engine, using the connection as a source of
database connectivity.
"""

from logging.config import fileConfig
from typing import Literal

from alembic import context
from alembic.util.exc import CommandError

from sqlalchemy.engine import Connection

from rechu.database import Database
from rechu.models import Base
from rechu.settings import Settings

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# TARGET_METADATA = mymodel.Base.metadata
TARGET_METADATA = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.

def run_migrations_offline() -> None:
    """
    Run migrations in 'offline' mode.

    This configures the context with just a URL and not an Engine, though an
    Engine is acceptable here as well. By skipping the Engine creation we don't
    even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the script output.
    """

    database = Database()
    if database.engine.name == 'sqlite':
        raise CommandError('Offline mode not supported for SQLite')

    context.configure(
        url=database.engine.url,
        target_metadata=TARGET_METADATA,
        render_as_batch=True,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"}
    )

    with context.begin_transaction():
        context.run_migrations()

def run_migrations_online() -> None:
    """
    Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.
    """

    database = Database()
    connectable = database.engine
    with connectable.connect() as connection:
        # SQLite batch operations sometimes drop tables to recreate them,
        # causing foreign key constraint violations
        if database.engine.name == 'sqlite':
            _set_sqlite_pragma(connection, "OFF")

        context.configure(
            connection=connection,
            target_metadata=TARGET_METADATA,
            render_as_batch=True
        )

        with context.begin_transaction():
            context.run_migrations()

        settings = Settings.get_settings()
        if database.engine.name == 'sqlite' and \
            settings.get('database', 'foreign_keys').lower() != 'off':
            _set_sqlite_pragma(connection, "ON")

def _set_sqlite_pragma(connection: Connection,
                       pragma_value: Literal["ON", "OFF"]) -> None:
    cursor = connection.connection.cursor()
    cursor.execute(f"PRAGMA foreign_keys = {pragma_value}")
    cursor.close()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""
Product metadata matcher.
"""

from collections.abc import Collection, Hashable, Iterator, Set
from enum import Enum
import logging
from typing import Optional
from sqlalchemy import and_, or_, cast, literal, select, Select, String
from sqlalchemy.orm import aliased, Session
from sqlalchemy.sql.expression import extract
from sqlalchemy.sql.functions import coalesce
from .base import Matcher
from ..models.base import Quantity
from ..models.product import Product, LabelMatch, PriceMatch, DiscountMatch
from ..models.receipt import Receipt, ProductItem, Discount, DiscountItems

LOGGER = logging.getLogger(__name__)

class MapKey(Enum):
    """
    Keys for a map of products with unique matchers and identifiers.
    """

    MAP_MATCH = 'match'
    MAP_SKU = 'sku'
    MAP_GTIN = 'gtin'

_MAP_KEYS = frozenset({MapKey.MAP_MATCH, MapKey.MAP_SKU, MapKey.MAP_GTIN})

class ProductMatcher(Matcher[ProductItem, Product]):
    """
    Matcher for receipt product items and product metadata.
    """

    IND_MINIMUM = 'minimum'
    IND_MAXIMUM = 'maximum'

    def __init__(self, map_keys: Set[MapKey] = _MAP_KEYS) -> None:
        super().__init__()
        self.discounts = True
        self._map_keys = map_keys

    def _get_specificity(self, product: Product) -> tuple[int, ...]:
        # A product has higher specificity if it has more of the three types
        # of matcher fields (label/price/discount) than another product, or if
        # theyboth have the same of these, then it is preferred if it has fewer
        # of those fields.
        matchers = bool(product.labels) + bool(product.prices)
        matcher_fields = len(product.labels) + len(product.prices)
        if self.discounts:
            matchers += bool(product.discounts)
            matcher_fields += len(product.discounts)
        return (matchers, -matcher_fields)

    def _select_generic(self, generic: Product, sub_range: Product) -> Product:
        if self._get_specificity(generic) >= self._get_specificity(sub_range):
            return generic

        return sub_range

    def select_duplicate(self, candidate: Product,
                         duplicate: Optional[Product]) -> Optional[Product]:
        if duplicate is not None:
            if candidate.id is not None and candidate.id == duplicate.id:
                return candidate
            if candidate.generic == duplicate:
                return self._select_generic(duplicate, candidate)
            if duplicate.generic == candidate:
                return self._select_generic(candidate, duplicate)
        return super().select_duplicate(candidate, duplicate)

    def _propose(self, product: Product,
                 item: ProductItem) -> Iterator[tuple[Product, ProductItem]]:
        if self.match(product, item):
            yield product, item

    def _find_dirty_candidates(self, session: Session,
                               items: Collection[ProductItem],
                               extra: Collection[Product],
                               only_unmatched: bool = False) \
            -> Iterator[tuple[Product, ProductItem]]:
        products = \
            session.scalars(select(Product)
                            .order_by(Product.generic_id.asc().nulls_first(),
                                      Product.id)).all()
        for item in items:
            if only_unmatched and item.product_id is not None:
                continue
            for product in products:
                yield from self._propose(product, item)
            for generic in extra:
                yield from self._propose(generic, item)
                for product_range in generic.range:
                    yield from self._propose(product_range, item)

    def find_candidates(self, session: Session,
                        items: Collection[ProductItem] = (),
                        extra: Collection[Product] = (),
                        only_unmatched: bool = False) \
            -> Iterator[tuple[Product, ProductItem]]:
        if any(item.id is None or item in session.dirty for item in items):
            yield from self._find_dirty_candidates(session, items, extra,
                                                   only_unmatched)
            return

        query = self._build_query(items, extra, only_unmatched)
        LOGGER.debug('%s', query)
        seen = set()
        for row in session.execute(query):
            if row.Product is not None:
                yield from self._propose(row.Product, row.ProductItem)
            if extra and row.ProductItem not in seen:
                seen.add(row.ProductItem)
                for product in extra:
                    yield from self._propose(product, row.ProductItem)
                    for product_range in product.range:
                        yield from self._propose(product_range, row.ProductItem)

    def _build_query(self, items: Collection[ProductItem],
                     extra: Collection[Product],
                     only_unmatched: bool) -> Select:
        minimum = aliased(PriceMatch)
        maximum = aliased(PriceMatch)
        other = aliased(PriceMatch)
        item_join = and_(ProductItem.label == coalesce(LabelMatch.name,
                                                       ProductItem.label),
                         ProductItem.price == coalesce(other.value *
                                                       ProductItem.amount,
                                                       ProductItem.price),
                         ProductItem.price.between(coalesce(minimum.value *
                                                            ProductItem.amount,
                                                            ProductItem.price),
                                                   coalesce(maximum.value *
                                                            ProductItem.amount,
                                                            ProductItem.price)))
        price_join = or_(other.value.is_(None),
                         ProductItem.unit.is_not_distinct_from(other.indicator),
                         other.indicator == cast(extract("year", Receipt.date),
                                                 String))
        query = select(ProductItem, Product)
        if extra:
            query = query.select_from(ProductItem) \
                .join(Product, literal(value=True), isouter=True) \
                .filter(item_join)
        else:
            query = query.select_from(Product)
        query = query.join(LabelMatch, Product.labels, isouter=True) \
            .join(other, Product.prices.and_(coalesce(other.indicator, '')
                                             .notin_((self.IND_MINIMUM,
                                                      self.IND_MAXIMUM))),
                  isouter=True) \
            .join(minimum,
                  Product.prices.and_(minimum.indicator == self.IND_MINIMUM),
                  isouter=True) \
            .join(maximum,
                  Product.prices.and_(maximum.indicator == self.IND_MAXIMUM),
                  isouter=True)
        if not extra:
            query = query.join(ProductItem, item_join)
        query = query.join(Receipt, ProductItem.receipt
                           .and_(Receipt.shop == coalesce(Product.shop,
                                                          Receipt.shop))
                           .and_(price_join))
        if self.discounts:
            discount_join = and_(Discount.id == DiscountItems.c.discount_id,
                                 Discount.label == coalesce(DiscountMatch.label,
                                                            Discount.label))
            query = query.join(DiscountMatch, Product.discounts, isouter=True) \
                .join(DiscountItems,
                      ProductItem.id == DiscountItems.c.product_id,
                      isouter=True) \
                .join(Discount, discount_join, isouter=True)
        if items:
            query = query \
                .filter(ProductItem.id.in_((item.id for item in items)))
        if only_unmatched:
            query = query.filter(ProductItem.product_id.is_(None))
        query = query.order_by(ProductItem.id,
                               Product.generic_id.asc().nulls_first(),
                               Product.id)

        return query

    @classmethod
    def _match_price(cls, price: PriceMatch, item: ProductItem) -> int:
        if item.quantity.unit is not None:
            try:
                quantity = Quantity(price.value, unit=f"1 / {price.indicator}")
                if quantity * item.quantity == item.price:
                    return 2
            except ValueError:
                pass

            return 0

        match_price = Quantity(price.value) * item.quantity
        if (price.indicator == cls.IND_MINIMUM and match_price <= item.price) \
            or \
           (price.indicator == cls.IND_MAXIMUM and match_price >= item.price):
            return 1
        if (price.indicator is None or
            price.indicator == str(item.receipt.date.year)) and \
            match_price == item.price:
            return 2

        return 0

    def match(self, candidate: Product, item: ProductItem) -> bool:
        # Candidate must be from the same shop and have at least one matcher
        # Currently, candidate must be generic instead of from a product range
        if candidate.shop != item.receipt.shop or (
                not candidate.labels and not candidate.prices and
                not candidate.discounts
            ):
            return False

        # One label matcher (if existing) must be the same as item label.
        if candidate.labels and \
            all(label.name != item.label for label in candidate.labels):
            return False

        seen_price = 0
        for price in candidate.prices:
            seen_price += self._match_price(price, item)
        # Must adhere to both 'minimum' and 'maximum', one date indicator,
        # one unit indicator or one price with no indicator. No price matchers
        # is also acceptable.
        if candidate.prices and seen_price < 2:
            return False

        # Final match check with discounts, one matching discount is enough.
        # No discount matcher is accepted, and so is an item without discounts
        # when the discount matching mode is disabled.
        if not candidate.discounts or not (self.discounts or item.discounts):
            return True
        for discount in candidate.discounts:
            if any(discount.label == bonus.label for bonus in item.discounts):
                return True

        return False

    @staticmethod
    def _get_product_match(product: Product) -> Optional[Hashable]:
        if not product.labels and not product.prices and not product.discounts:
            return None
        return (
            product.shop,
            tuple(label.name for label in product.labels),
            tuple(
                (price.indicator, price.value)
                for price in product.prices
            ),
            tuple(discount.label for discount in product.discounts)
        )

    def _get_keys(self, product: Product) -> tuple[tuple[Hashable, ...], ...]:
        keys = (
            (MapKey.MAP_MATCH, self._get_product_match(product)),
            (MapKey.MAP_SKU, product.shop, product.sku),
            (MapKey.MAP_GTIN, product.shop, product.gtin)
        )
        return tuple(key for key in keys if key[0] in self._map_keys)

    def load_map(self, session: Session) -> None:
        self._map = {}
        query = select(Product).order_by(Product.generic_id.asc().nulls_first(),
                                         Product.id)
        for product in session.scalars(query):
            self.add_map(product)

    def add_map(self, candidate: Product) -> bool:
        if self._map is None:
            return False

        add = False
        for key in self._get_keys(candidate):
            if key[-1] is not None:
                add = self._map.setdefault(key, candidate) is candidate or add

        for product_range in candidate.range:
            add = self.add_map(product_range) or add

        return add

    def discard_map(self, candidate: Product) -> bool:
        if self._map is None:
            return False

        remove = False
        for key in self._get_keys(candidate):
            product = self._map.pop(key, None)
            if product is candidate:
                remove = True
            elif product is not None:
                LOGGER.warning('Product instance stored at %r is not %r: %r',
                               key, candidate, product)
                self._map[key] = product

        for product_range in candidate.range:
            remove = self.discard_map(product_range) or remove

        return remove

    def check_map(self, candidate: Product) -> Optional[Product]:
        if self._map is None:
            return None
        for key in self._get_keys(candidate):
            if key in self._map:
                return self._map[key]
        return None



"""
Models for receipt data.
"""

import datetime
from typing import Optional
from sqlalchemy import Column, ForeignKey, String, Table
from sqlalchemy.orm import MappedColumn, Relationship,  mapped_column, \
    relationship
from .base import Base, Price, Quantity, Unit
from .product import Product

class Receipt(Base): # pylint: disable=too-few-public-methods
    """
    Receipt model for a receipt from a certain date at a shop with products
    and possibly discounts.
    """

    __tablename__ = "receipt"

    filename: MappedColumn[str] = mapped_column(String(255), primary_key=True)
    updated: MappedColumn[datetime.datetime]
    date: MappedColumn[datetime.date]
    shop: MappedColumn[str] = mapped_column(String(32)) # shop.key
    products: Relationship[list["ProductItem"]] = \
        relationship(back_populates="receipt", cascade="all, delete-orphan",
                     passive_deletes=True, order_by="ProductItem.position")
    discounts: Relationship[list["Discount"]] = \
        relationship(back_populates="receipt", cascade="all, delete-orphan",
                     passive_deletes=True, order_by="Discount.position")

    @property
    def total_price(self) -> Price:
        """
        Retrieve the total cost of the receipt after discounts.
        """

        total = sum(product.price for product in self.products) + \
            sum(discount.price_decrease for discount in self.discounts)
        return Price(total)

    def __repr__(self) -> str:
        return f"Receipt(date={self.date.isoformat()!r}, shop={self.shop!r})"

# Association table for products involved in discounts
DiscountItems = Table("receipt_discount_products", Base.metadata,
                      Column("discount_id", ForeignKey('receipt_discount.id',
                                                       ondelete='CASCADE'),
                             primary_key=True),
                      Column("product_id", ForeignKey('receipt_product.id',
                                                      ondelete='CASCADE'),
                             primary_key=True))

class ProductItem(Base): # pylint: disable=too-few-public-methods
    """
    Product model for a product item mentioned on a receipt.
    """

    __tablename__ = "receipt_product"

    id: MappedColumn[int] = mapped_column(primary_key=True)
    receipt_key: MappedColumn[str] = \
        mapped_column(ForeignKey('receipt.filename', ondelete='CASCADE'))
    receipt: Relationship[Receipt] = relationship(back_populates="products")

    quantity: MappedColumn[Quantity]
    label: MappedColumn[str]
    price: MappedColumn[Price]
    discount_indicator: MappedColumn[Optional[str]]
    discounts: Relationship[list["Discount"]] = \
        relationship(secondary=DiscountItems, back_populates="items",
                     passive_deletes=True)
    product_id: MappedColumn[Optional[int]] = \
        mapped_column(ForeignKey('product.id', ondelete='SET NULL'))
    product: Relationship[Optional[Product]] = relationship()
    position: MappedColumn[int]
    # Extracted fields from quantity
    amount: MappedColumn[float]
    unit: MappedColumn[Optional[Unit]]

    def __repr__(self) -> str:
        return (f"ProductItem(receipt={self.receipt_key!r}, "
                f"quantity='{self.quantity!s}', label={self.label!r}, "
                f"price={self.price!s}, "
                f"discount_indicator={self.discount_indicator!r}, "
                f"product={self.product_id!r})")

class Discount(Base): # pylint: disable=too-few-public-methods
    """
    Discount model for a discount action mentioned on a receipt.
    """

    __tablename__ = "receipt_discount"

    id: MappedColumn[int] = mapped_column(primary_key=True)
    receipt_key: MappedColumn[str] = \
        mapped_column(ForeignKey('receipt.filename', ondelete='CASCADE'))
    receipt: Relationship[Receipt] = relationship(back_populates="discounts")

    label: MappedColumn[str]
    price_decrease: MappedColumn[Price]
    items: Relationship[list[ProductItem]] = \
        relationship(secondary=DiscountItems, back_populates="discounts",
                     passive_deletes=True)
    position: MappedColumn[int]

    def __repr__(self) -> str:
        return (f"Discount(receipt={self.receipt_key!r}, label={self.label!r}, "
                f"price_decrease={self.price_decrease!s}, "
                f"items={[item.label for item in self.items]!r})")



"""
Database entity matching methods.
"""

from collections.abc import Collection, Hashable, Iterable, Iterator
from typing import Generic, Optional, TypeVar
from sqlalchemy.orm import Session
from ..inventory.base import Inventory
from ..models.base import Base as ModelBase

IT = TypeVar('IT', bound=ModelBase)
CT = TypeVar('CT', bound=ModelBase)

class Matcher(Generic[IT, CT]):
    """
    Generic item candidate model matcher.
    """

    def __init__(self) -> None:
        self._map: Optional[dict[Hashable, CT]] = None

    def find_candidates(self, session: Session,
                        items: Collection[IT] = (),
                        extra: Collection[CT] = (),
                        only_unmatched: bool = False) \
            -> Iterator[tuple[CT, IT]]:
        """
        Detect candidate models in the database that match items. Optionally,
        the `items` may be provided, which might not have been inserted or
        updated in the database, otherwise all items from the database are
        attempted for matching. Moreover, `extra` candidates may be provided,
        which in their case augment those from the database. If `only_unmatched`
        is enabled, then only items that do not have a relation with a candidate
        model are attempted for matching. The resulting iterator provides tuples
        of matches between candidates and items which have not had an update to
        their match relationship yet; multiple candidate models may be indicated
        for a single item model.
        """

        raise NotImplementedError('Search must be implemented by subclasses')

    def filter_duplicate_candidates(self, candidates: Iterable[tuple[CT, IT]]) \
            -> Iterator[tuple[CT, IT]]:
        """
        Detect if item models were matched against multiple candidates and
        filter out such models.
        """

        seen: dict[IT, Optional[CT]] = {}
        for candidate, item in candidates:
            if item in seen:
                seen[item] = self.select_duplicate(candidate, seen[item])
            else:
                seen[item] = candidate
        for item, unique in seen.items():
            if unique is not None:
                yield unique, item

    def select_duplicate(self, candidate: CT, duplicate: Optional[CT]) \
            -> Optional[CT]: # pylint: disable=unused-argument
        """
        Determine which of two candidate models should be matched against some
        item, if any. If this returns `None` than neither of the models is
        provided as a match.
        """

        if candidate is duplicate:
            return candidate

        return None

    def match(self, candidate: CT, item: IT) -> bool:
        """
        Check if a candidate model matches an item model without looking up
        through the database.
        """

        raise NotImplementedError('Match must be implemented by subclasses')

    def load_map(self, session: Session) -> None:
        """
        Create a mapping of unique keys of candidate models to their database
        entities.
        """
        # pylint: disable=unused-argument

        self._map = {}

    def clear_map(self) -> None:
        """
        Clear the mapping of unique keys of candidate models to entities
        such that it no database entities are matched.
        """

        self._map = {}

    def fill_map(self, inventory: Inventory[CT]) -> None:
        """
        Update a mapping of unique keys of candidate models from a filled
        inventory.
        """

        if self._map is None:
            self._map = {}
        for group in inventory.values():
            for model in group:
                self.add_map(model)

    def add_map(self, candidate: CT) -> bool:
        """
        Manually add a candidate model to a mapping of unique keys. Returns
        whether the entity was actually added, which is not done if the map is
        not initialized or the keys are not unique enough.
        """
        # pylint: disable=unused-argument

        return False

    def discard_map(self, candidate: CT) -> bool:
        """
        Remove a candidate model to a mapping of unique keys. Returns whether
        the entity was actually removed.
        """
        # pylint: disable=unused-argument

        return False

    def check_map(self, candidate: CT) -> Optional[CT]:
        """
        Retrieve a candidate model obtained from the database which has one or
        more of the unique keys in common with the provided `candidate`. If no
        such candidate is found, then `None` is returned. Any returned candidate
        should be considered read-only due to it coming from an earlier session
        that is already closed.
        """
        # pylint: disable=unused-argument

        return None



"""
Database schema creation subcommand.
"""

from .base import Base
from ..database import Database

@Base.register("create")
class Create(Base):
    """
    Create the database with the database schema.
    """

    subparser_keywords = {
        'help': 'Create the database and schema',
        'description': 'Create database schema tables at the configured URI.'
    }

    def run(self) -> None:
        database = Database()
        database.create_schema()



"""
Command entry point.
"""

import sys
from .command.base import Base

def main() -> None:
    """
    Main entry point for receipt subcommands.
    """

    Base.start(sys.executable, sys.argv)

if __name__ == "__main__":
    main()



"""
Subcommand to run Alembic commands for database migration.
"""

from alembic.config import CommandLine
from .base import Base
from ..database import Database

@Base.register("alembic")
class Alembic(Base):
    """
    Run an alembic command.
    """

    subparser_keywords = {
        # Let alembic handle `rechu alembic --help` argument
        'add_help': False,
        # Describe command in `rechu --help`
        'help': 'Perform database revision management',
        # Pass along all arguments to alembic even if they start with dashes
        'prefix_chars': '\x00'
    }
    subparser_arguments = [
        ('args', {'nargs': '*', 'help': 'alembic arguments'})
    ]

    def __init__(self) -> None:
        super().__init__()
        self.args: list[str] = []

    def run(self) -> None:
        alembic_config = Database.get_alembic_config()

        alembic_cmd = CommandLine(prog=f"{self.program} {self.subcommand}")

        subcommand = self.args[0] if self.args else ""
        args: list[str] = []
        if subcommand != "":
            args.append(subcommand)
        if subcommand == "revision":
            args.append("--autogenerate")
        args.extend(self.args[1:])

        alembic_options = alembic_cmd.parser.parse_args(args)
        alembic_cmd.run_cmd(alembic_config, alembic_options)



"""
Subcommand to import receipt YAML files.
"""

from collections.abc import Hashable
from datetime import datetime
from itertools import chain
import re
from pathlib import Path
from sqlalchemy import select
from sqlalchemy.orm import Session
from .base import Base
from ..database import Database
from ..inventory.products import Products
from ..io.products import ProductsReader
from ..io.receipt import ReceiptReader
from ..matcher.product import ProductMatcher
from ..models import Receipt

_ProductMap = dict[str, dict[Hashable, int]]

def get_updated_time(path: Path) -> datetime:
    """
    Retrieve the latest modification time of a file or directory in the `path`
    as a `datetime` object.
    """

    return datetime.fromtimestamp(path.stat().st_mtime)

@Base.register("read")
class Read(Base):
    """
    Read updated YAML files and import them to the database.
    """

    subparser_keywords = {
        'help': 'Import updated product and receipt files to the database',
        'description': 'Find YAML files for products and receipts stored in '
                       'the data paths and import new or updated entities to '
                       'the database.'
    }
    def run(self) -> None:
        data_path = Path(self.settings.get('data', 'path'))


        with Database() as session:
            _, products_glob, _, products_pattern = \
                Products.get_parts(self.settings)
            self._handle_products(session, data_path, products_glob)
            session.flush()
            new_receipts = self._handle_receipts(session, data_path,
                                                 products_pattern)
            self._update_matches(session, new_receipts)

    def _handle_products(self, session: Session, data_path: Path,
                         products_glob: str) -> None:
        matcher = ProductMatcher()
        products = Products.select(session)
        matcher.fill_map(products)
        unseen = set(chain(*products.values()))

        for path in sorted(data_path.glob(products_glob)):
            self.logger.info('Looking at products in %s', path)
            try:
                for product in ProductsReader(path).read():
                    existing = matcher.check_map(product)
                    if existing is None:
                        session.add(product)
                    else:
                        product.id = existing.id
                        unseen.discard(existing)
                        session.merge(product)
            except (TypeError, ValueError):
                self.logger.exception('Could not parse product from %s', path)

        for removed in unseen:
            self.logger.warning('Deleting %r from database', removed)
            session.delete(removed)

    def _handle_receipts(self, session: Session, data_path: Path,
                         products_pattern: re.Pattern[str]) -> list[Receipt]:
        data_pattern = self.settings.get('data', 'pattern')

        receipts = {
            receipt.filename: receipt.updated
            for receipt in session.scalars(select(Receipt))
        }
        new_receipts: list[Receipt] = []

        latest = max(receipts.values(), default=datetime.min)
        directories = [data_path] if data_pattern == '.' else \
            data_path.glob(data_pattern)
        self.logger.info('Latest update timestamp in DB: %s', latest)
        for data_directory in directories:
            # Look at directories with recent files (not strictly updated
            # because multiple files may have the same updated time)
            if data_directory.is_dir() and \
                get_updated_time(data_directory) >= latest:
                self.logger.info('Looking at files in %s', data_directory)
                new_receipts.extend(self._handle_directory(data_directory,
                                                           receipts, session,
                                                           products_pattern))

        return new_receipts

    def _handle_directory(self, data_directory: Path,
                          receipts: dict[str, datetime],
                          session: Session,
                          products_pattern: re.Pattern) -> list[Receipt]:
        new_receipts: list[Receipt] = []
        for path in data_directory.glob('*.yml'):
            if products_pattern.match(str(path)):
                continue
            if self._is_updated(path, receipts):
                try:
                    receipt = next(ReceiptReader(path,
                                                 updated=datetime.now()).read())
                    if receipt.filename in receipts:
                        receipt = session.merge(receipt)
                    else:
                        session.add(receipt)
                    new_receipts.append(receipt)
                except (StopIteration, TypeError):
                    self.logger.exception('Could not retrieve receipt %s', path)

        return new_receipts

    @staticmethod
    def _is_updated(receipt_path: Path, receipts: dict[str, datetime]) -> bool:
        if receipt_path.name not in receipts:
            return True

        updated = get_updated_time(receipt_path)
        # Check if the updated time is stricly newer
        return updated > receipts[receipt_path.name]

    def _update_matches(self, session: Session,
                        receipts: list[Receipt]) -> None:
        matcher = ProductMatcher()
        items = list(chain(*(receipt.products for receipt in receipts)))
        pairs = matcher.find_candidates(session, items, only_unmatched=True)
        for product, item in matcher.filter_duplicate_candidates(pairs):
            self.logger.info('Matching %r with %r', item, product)
            item.product = product



"""
Database access.
"""

from pathlib import Path
import sys
from types import TracebackType
from typing import Optional, TextIO
from alembic.config import Config
from alembic import script
from alembic.runtime.migration import MigrationContext
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.engine.interfaces import DBAPIConnection
from sqlalchemy.pool import ConnectionPoolEntry
from sqlalchemy.orm import Session
from .models import Base
from .settings import Settings

class Database:
    """
    Database provider.
    """

    def __init__(self) -> None:
        settings = Settings.get_settings()
        self.engine = create_engine(settings.get('database', 'uri'))
        self.session: Optional[Session] = None

        if self.engine.name == 'sqlite':
            event.listen(Engine, "connect", self._set_sqlite_pragma)

    @staticmethod
    def _set_sqlite_pragma(connection: DBAPIConnection,
                           _connection_record: ConnectionPoolEntry) -> None:
        settings = Settings.get_settings()
        pragma_value = "OFF" \
            if settings.get('database', 'foreign_keys').lower() == 'off' \
            else "ON"
        cursor = connection.cursor()
        cursor.execute(f"PRAGMA foreign_keys = {pragma_value}")
        cursor.close()

    def create_schema(self) -> None:
        """
        Perform schema creation on an empty database, marking it as up to date
        using alembic's stamp command.
        """

        Base.metadata.create_all(self.engine)

        alembic_config = self.get_alembic_config()
        directory = script.ScriptDirectory.from_config(alembic_config)
        with self.engine.begin() as connection:
            migration_context = MigrationContext.configure(connection)
            migration_context.stamp(directory, "head")

    def drop_schema(self) -> None:
        """
        Clean up the database by removing all model tables.
        """

        Base.metadata.drop_all(self.engine)

    @staticmethod
    def get_alembic_config(stdout: TextIO = sys.stdout) -> Config:
        """
        Retrieve the configuration object for alembic preconfigured for rechu.
        """

        package_root = Path(__file__).resolve().parent
        return Config(package_root / 'alembic.ini', stdout=stdout)

    def __del__(self) -> None:
        self.close()

    def __enter__(self) -> Session:
        if self.session is not None:
            raise RuntimeError('Detected nested database session attempts')
        self.session = Session(self.engine)
        return self.session

    def __exit__(self, exc_type: Optional[type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        if self.session is not None:
            self.session.commit()
        self.close()

    def close(self) -> None:
        """
        Close any open database session connection.
        """

        if self.session is not None:
            self.session.close()
            self.session = None

    def clear(self) -> None:
        """
        Reset any permanent settings-based state of the database.
        """

        if event.contains(Engine, "connect", self._set_sqlite_pragma):
            event.remove(Engine, "connect", self._set_sqlite_pragma)



"""
Attribute types for measurable values such as quantities and units.
"""

from .decorator import QuantityType, UnitType
from .quantity import Quantity
from .unit import Unit

__all__ = ["Quantity", "Unit", "QuantityType", "UnitType"]



"""
Add receipt item position fields

Revision ID: b1a7a91c8de8
Revises: 3b0cfa853967
Create Date: 2025-05-08 23:06:02.101619
"""
# pylint: disable=invalid-name

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column


# Revision identifiers, used by Alembic.
revision: str = 'b1a7a91c8de8'
down_revision: Union[str, None] = '3b0cfa853967'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """
    Perform the upgrade.
    """

    with op.batch_alter_table('receipt_discount', schema=None) as batch_op:
        batch_op.add_column(sa.Column('position', sa.Integer(), nullable=True))
        discount = table('receipt_discount', column('id', sa.Integer()),
                         column('position', sa.Integer()))
    with op.batch_alter_table('receipt_discount', schema=None) as batch_op:
        batch_op.execute(discount.update().values(position=discount.c.id))
        batch_op.alter_column('position',
                              existing_type=sa.Integer(),
                              existing_nullable=True,
                              nullable=False)

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        batch_op.add_column(sa.Column('position', sa.Integer(), nullable=True))
        product = table('receipt_product', column('id', sa.Integer()),
                         column('position', sa.Integer()))
    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        batch_op.execute(product.update().values(position=product.c.id))
        batch_op.alter_column('position',
                              existing_type=sa.Integer(),
                              existing_nullable=True,
                              nullable=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    """
    Perform the downgrade.
    """

    with op.batch_alter_table('receipt_product', schema=None) as batch_op:
        batch_op.drop_column('position')

    with op.batch_alter_table('receipt_discount', schema=None) as batch_op:
        batch_op.drop_column('position')

    # ### end Alembic commands ###



"""
Models for shop metadata.
"""

from typing import Optional
from sqlalchemy import String
from sqlalchemy.orm import MappedColumn, mapped_column
from .base import Base

class Shop(Base): # pylint: disable=too-few-public-methods
    """
    Shop metadata model.
    """

    __tablename__ = "shop"

    key: MappedColumn[str] = mapped_column(String(32), primary_key=True)
    name: MappedColumn[Optional[str]] = mapped_column(String(32))
    website: MappedColumn[Optional[str]]
    wikidata: MappedColumn[Optional[str]]

    def __repr__(self) -> str:
        return (f"Shop(key={self.key!r}, name={self.name!r}, "
                f"website={self.website!r}, wikidata={self.wikidata!r})")



"""
Subcommand to export database entries as YAML files.
"""

from pathlib import Path
from typing import TypeVar
from sqlalchemy import select
from sqlalchemy.orm import Session
from .base import Base
from ..database import Database
from ..inventory.base import Selectors
from ..inventory.products import Products
from ..io.base import Writer
from ..io.receipt import ReceiptWriter
from ..models import Base as ModelBase, Receipt

T = TypeVar('T', bound=ModelBase)

@Base.register("dump")
class Dump(Base):
    """
    Dump YAML files from the database.
    """

    subparser_keywords = {
        'help': 'Export entities from the database',
        'description': 'Create one or more YAML files for data in the database.'
    }
    subparser_arguments = [
        ('files', {
            'metavar': 'FILE',
            'nargs': '*',
            'help': 'One or more product inventories or receipts to write; if '
                    'no filenames are given, then the entire database is dumped'
        })
    ]

    def __init__(self) -> None:
        super().__init__()
        self.files: list[str] = []
        self.data_path = Path(self.settings.get('data', 'path'))
        self._directories: set[Path] = set()

    def run(self) -> None:
        products_pattern = Products.get_parts(self.settings)[-1]

        products_files: Selectors = []
        receipt_files: list[str] = []
        for file in self.files:
            products_match = products_pattern.match(file)
            if products_match:
                products_files.append(products_match.groupdict())
            else:
                # Filter off path elements to just keep the file name
                receipt_files.append(Path(file).name)

        with Database() as session:
            self._write_products(session, products_files)
            self._write_receipts(session, receipt_files)

    def _write_products(self, session: Session, files: Selectors) -> None:
        for writer in Products.select(session, selectors=files).get_writers():
            self._write(writer)

    def _write_receipts(self, session: Session, files: list[str]) -> None:
        data_format = self.settings.get('data', 'format')

        receipts = select(Receipt)
        if files:
            receipts = receipts.where(Receipt.filename.in_(files))
        for receipt in session.scalars(receipts):
            path_format = self.data_path / data_format.format(date=receipt.date,
                                                              shop=receipt.shop)
            path = path_format.parent / receipt.filename
            self._write(ReceiptWriter(path, (receipt,)))

    def _write(self, writer: Writer[T]) -> None:
        # Only write new files, do not overwrite existing ones
        path = writer.path
        if not path.exists():
            if path.parent not in self._directories:
                # Create directories when needed, cache directories
                path.parent.mkdir(parents=True, exist_ok=True)
                self._directories.add(path.parent)

            writer.write()
